/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku;

import java.awt.Point;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Random;

import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.Puzzle.PuzzleDifficulty;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.solver.BruteForceSolver;

/**
 * Generates a sudoku grid
 * 
 * @author A.G.Docherty
 *
 */
public class Generator
{
private static final List<Integer> allPossibleNumbers = new ArrayList<Integer>();
private static final int MAX_ATTEMPTS = 20;
public static final int MAX_QUEUE_SIZE = 10;

static
	{
	for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
	    {
	    allPossibleNumbers.add(i+1);
	    }
	}

private final List<StoreListener> storeListeners = new ArrayList<StoreListener>();
private final Puzzle puzzle;
private final Random random = new Random(System.currentTimeMillis());
private int iterations;
private BruteForceSolver bruteForce;
private final StoreFiller storeFiller;
private volatile boolean cancelled;
private final SudokuSolver solver;

Generator()
	{
	puzzle = new Puzzle();
	solver = new SudokuSolver();
	solver.setPuzzle(puzzle);
	
	storeFiller = new StoreFiller();
	}

/**
 * Starts the background generator thread. As puzzles are generated they are placed into a
 * store for quick retrieval.
 */
public void startBackgroundGeneration()
	{
	storeFiller.start();
	}

/**
 * Stops the background generator thread. The store is not emptied of previously generated puzzles.
 */
public void stopBackgroundGeneration()
	{
	storeFiller.stop();
	}

/**
 * Generates a puzzle with this difficulty level
 *  
 * @param difficulty the difficulty level of the puzzle to generate
 * @param listener a listener of progress events. For each new attempt an event is sent. 
 * @return the puzzle or null if cancelled
 */
synchronized Puzzle generatePuzzle(PuzzleDifficulty difficulty, PropertyChangeListener listener)
	{
    PuzzleStoreItem item = storeFiller.getPuzzle(difficulty);
    
    if ( item == null )
		{
		storeFiller.pause(true);
		item = getNewPuzzle(difficulty, listener);
		}
	
	storeFiller.pause(false);

	if ( item != null )
		return item.puzzle;
	
	return null;
	}

/**
 * Generates a puzzle with this difficulty level
 *  
 * @param difficulty
 * @param listener 
 * @return the puzzle or null if cancelled
 */
private synchronized PuzzleStoreItem getNewPuzzle(PuzzleDifficulty difficulty, PropertyChangeListener listener)
	{
	Puzzle p = new Puzzle();
	bruteForce = new BruteForceSolver(p, random);

	cancelled = false;
	int iteration = 0;
	int completed = 0;
	do 
		{
		p.clear();
		
		// fill a block
		int added = bruteForce.fillBlock(0);
//		added += bruteForce.fillBlock(8);
//		added += bruteForce.fillBlock(4);
		
		try
	        {
	        completed = bruteForce.findSolutions();
	        }
        catch ( IllegalGridException exc )
	        {
	        // this shouldn't happen as the brute force solver should always find a solution
	        continue;
	        }
		
		if ( completed == SudokuSolver.GRID_SIZE * SudokuSolver.GRID_SIZE - added )
			{
			PuzzleStoreItem psItem = null;
			psItem = makePuzzle(p, difficulty);
			
			if ( psItem != null )
				{
//			    System.out.println("Generated puzzle, is legal "+p.isLegal());
//				if ( !psItem.puzzle.isUnique() )
//					System.out.println("WARN: this puzzle does not have a unique solution");
				
				return psItem;
				}
			}
		
		if ( listener != null )
			{
			listener.propertyChange(new PropertyChangeEvent(this, "progress", iteration, ++iteration));
			}
		}
	while ( !cancelled );
	
	return null;
	}

/**
 * Cancels any ongoing generation
 */
public void cancelGenerate()
	{
	cancelled = true;
	}


//works in reverse - each loop reduces the number of known cells
//reduces the number of recreated object by reusing the same root grid
// fills map of possible values
/**
 * Makes a puzzle by randomly removing cells from a completed grid until the max allowable number 
 * of cells are left and seeing if the grid can be solved from these numbers. If it can 
 * one of the numbers is removed and solution is attempted. This step is repeated until
 * solution is no longer possible or the grid is already harder than required.
 * Once the minimum number of cells is found if the difficulty is correct the puzzle is returned.
 * else the above process is repeated up to MAX_ATTEMPTS times. If no puzzle has been found 
 */
private PuzzleStoreItem makePuzzle(Puzzle p, PuzzleDifficulty dif)
	{
	Cell[][] grid = p.getGrid();
	List<Point> locs = new ArrayList<Point>();
    Cell[][] newGrid = Puzzle.copyGrid(grid);

    int minCells;
    int maxCells;
    
    if ( dif == null )
    	{
        minCells = PuzzleDifficulty.EXTREME.minNumbers();
        maxCells = PuzzleDifficulty.EASY.maxNumbers();
    	}
    else
    	{
        minCells = dif.minNumbers();
        maxCells = dif.maxNumbers();
    	}
    
    int totalNumCells = SudokuSolver.GRID_SIZE * SudokuSolver.GRID_SIZE;
    
    // try removing values until we find something that is solvable with the minimum number of cells
    for ( int n = maxCells; n >= minCells; n-- )
    	{
    	Puzzle.copyGrid(grid, newGrid);
    	locs.clear();

		// if failed try up to MAX_ATTEMPTS other random combinations
    	for ( int i = 0; i < MAX_ATTEMPTS; i++ )
    		{
    		// for each location reinstate it's value 
    		for ( Point loc : locs )
    			{
    			newGrid[loc.y][loc.x].clear();
    			newGrid[loc.y][loc.x].addAll(grid[loc.y][loc.x]);
    			}

    		locs.clear();

    		int offset = 0;
    		PuzzleStoreItem foundPuzzle = null;

    		// select a new set of missing number
    		while ( true )
    			{
    			iterations++;
    			// generate some unique random points for the missing numbers
    			while ( locs.size() < totalNumCells - n + offset )
    				{
    				int x = random.nextInt(SudokuSolver.GRID_SIZE);
    				int y = random.nextInt(SudokuSolver.GRID_SIZE);

    				Point loc = new Point(x, y);

    				if ( !locs.contains(loc) )
    					{
    					newGrid[y][x].addAll(Puzzle.allPossibleNumbers);
    					locs.add(loc);
    					}
    				}

    			// set this grid as the one to solve
    			puzzle.setGrid(newGrid);

    			int difVal = 0;
    			try
    				{
    				// solve the puzzle
    				solver.setPuzzle(puzzle);
    				difVal = solver.compute();
    				}
    			catch ( IllegalGridException exc )
    				{
    				// not solvable so try another grid
    				difVal = -99;
    				}

				puzzle.setDifficulty(solver.getPuzzle().getDifficulty());

				// if this hasn't solved
    			if ( difVal <= 0 )
    				{
    				// if we previously had a solution use it 
    				if ( foundPuzzle != null )
    					return foundPuzzle;

    				break;
    				}

    			// if this puzzle is the correct difficulty keep a copy of it
    			if ( dif == null || puzzle.getDifficulty() == dif )
    				{
    				foundPuzzle = new PuzzleStoreItem(difVal, totalNumCells-locs.size(), puzzle.clone());
    				}
    			
    			// if this puzzle is harder than we need then don't try any more for
    			// this set up
    			if ( dif != null && puzzle.getDifficulty().ordinal() > dif.ordinal() )
    				break;

    			offset++;
    			}
    		}
    	}
	
	return null;
	}

/**
 * Adds a listener. 
 *
 * @param l - the listener to add
 */
public void addStoreListener(StoreListener l)
	{
	storeListeners.add(l);
	}

/**
 * Removes the listener
 *
 * @param l - the listener to remove
 */
public void removeStoreListener(StoreListener l)
	{
	storeListeners.remove(l);
	}

/**
 * Removes the listener
 */
public void removeAllStoreListeners()
	{
	storeListeners.clear();
	}

/**
 * Fires the store event to all listeners
 */
private void fireItemStored(PriorityQueue<PuzzleStoreItem> store, PuzzleStoreItem item)
	{
	// call all listeners
	for ( StoreListener listener : storeListeners )
		{
		listener.itemStored(store, item);
		}
	}

/**
 * Fires the store event to all listeners
 */
private void fireItemRemoved(PriorityQueue<PuzzleStoreItem> store, PuzzleStoreItem item)
	{
	// call all listeners
	for ( StoreListener listener : storeListeners )
		{
		listener.itemRetrieved(store, item);
		}
	}

/**
 * Saves the puzzles in this store
 * 
 * @throws FileNotFoundException 
 */
public void save() throws FileNotFoundException
	{
	storeFiller.write();
	}

/**
 * Loads the puzzles in this store
 * 
 * @throws IllegalGridException 
 * @throws IOException 
 */
public void load() throws IOException, IllegalGridException
	{
	storeFiller.read();
	}

/**
 * 
 * @author A.G.Docherty
 *
 */
public static class PuzzleStoreItem implements Comparable<PuzzleStoreItem>, Cloneable
{
private final int difficulty;
private Puzzle puzzle;
private final int nCells;

PuzzleStoreItem(int d, int c, Puzzle p)
	{
	difficulty = d;
	nCells = c;
	puzzle = p;
	}

/**
 * @return the difficulty
 */
public int getDifficulty()
	{
	return difficulty;
	}

/**
 * @return the puzzle
 */
public Puzzle getPuzzle()
	{
	return puzzle;
	}

@Override
public int compareTo(PuzzleStoreItem o)
	{
	if ( o == null )
		return 1;
	
	int d = difficulty - o.difficulty;
	
	if ( d == 0 )
		{
		return o.nCells - nCells;
		}
	
	return d;
	}

@Override
public boolean equals(Object obj)
	{
	if ( obj == null ||  obj.getClass() != this.getClass() )
		return false;
	
	PuzzleStoreItem psi = (PuzzleStoreItem)obj;
	
	if ( difficulty != psi.difficulty )
		return false;
	
	if ( nCells != psi.nCells )
		return false;

	return true;
	}

@Override
public int hashCode()
	{
	return (difficulty * 7)+(nCells*13);
	}

@Override
protected PuzzleStoreItem clone() 
    {
    PuzzleStoreItem obj;
    try
	    {
	    obj = (PuzzleStoreItem)super.clone();
	    }
    catch ( CloneNotSupportedException exc )
	    {
	    // This won't ever happen
	    exc.printStackTrace();
	    return null;
	    }
    
    obj.puzzle = puzzle.clone();
    
    return obj;
    }

@Override
public String toString()
	{
	return "Diff = "+difficulty+", initial num cells = "+nCells;
	}
}

/**
 * 
 * @author A.G.Docherty
 *
 */
private class StoreFiller implements Runnable
{
private static final String CELLS = "Cells";
private static final String DIFFICULTY = "Difficulty";
private final Map<PuzzleDifficulty, PriorityQueue<PuzzleStoreItem>> gridStore;
private final String filename = "GridStore.msdk";
private Thread bkgThread;
private transient boolean active;
private transient boolean paused;

StoreFiller()
	{
	active = false;
	paused = false;
	
	gridStore = new HashMap<PuzzleDifficulty, PriorityQueue<PuzzleStoreItem>>();
	gridStore.put(PuzzleDifficulty.BEGINNER, new PriorityQueue<PuzzleStoreItem>());
	gridStore.put(PuzzleDifficulty.EASY, new PriorityQueue<PuzzleStoreItem>());
	gridStore.put(PuzzleDifficulty.MEDIUM, new PriorityQueue<PuzzleStoreItem>());
	gridStore.put(PuzzleDifficulty.HARD, new PriorityQueue<PuzzleStoreItem>());
	gridStore.put(PuzzleDifficulty.EXTREME, new PriorityQueue<PuzzleStoreItem>());
	}

public PuzzleStoreItem getPuzzle(PuzzleDifficulty difficulty)
	{
	PriorityQueue<PuzzleStoreItem> store = gridStore.get(difficulty);
	
	PuzzleStoreItem item = null;
	if ( store.size() > 0 )
		{
		synchronized ( store )
	        {
	        // get the last item
	        for ( PuzzleStoreItem psi : store )
	            {
	            item = psi;
	            }
	        
	        // remove the item if we have found it
	        if ( item != null )
	        	{
	        	store.remove(item);
				fireItemRemoved(store, item);
	        	}
	        }
		}
	
	return item;
	}

public void pause(boolean b)
	{
	paused = b;
	}

void stop()
	{
	if ( bkgThread == null )
		return;
	
	active = false;
	bkgThread = null;
	}

void start()
	{
	if ( bkgThread != null )
		return;
	
	active = true;
	paused = false;
	
	bkgThread = new Thread(this, "SudokuStoreFiller");
	bkgThread.setDaemon(true);
	bkgThread.setPriority(Thread.NORM_PRIORITY-1);
	bkgThread.start();
	}

@Override
public void run()
	{
	while ( active )
		{
		while ( paused )
			{
			try
	            {
	            Thread.sleep(200);
	            }
            catch ( InterruptedException exc )
	            {
	            // do nothing - this will loop until paused is false
	            }
            
        	if ( !paused )
        		{
        		try
        	        {
        	        Thread.sleep(200);
        	        }
        	    catch ( InterruptedException exc )
        	        {
        	        // do nothing - this will loop until paused is false
        	        }
        		}
			}
		
		PuzzleStoreItem psi = getNewPuzzle(null, null);
		
		if ( psi == null )
			continue;

		// Check to see if this puzzle has a unique solution
//		if ( !psi.puzzle.isUnique() )
//			{
//			System.out.println("WARN: this puzzle does not have a unique solution");
//		    int count = psi.puzzle.makeUnique();
//			System.out.println("WARN: this puzzle was made unique by adding "+count);
//			}
		
		// how difficult is it
		addToStore(psi);
		}
	}

/**
 * Adds the puzzle store item to the store queue
 * 
 * @param psi
 * @param dif
 */
private void addToStore(PuzzleStoreItem psi)
	{
	PuzzleDifficulty dif = psi.puzzle.getDifficulty();
	PriorityQueue<PuzzleStoreItem> queue = gridStore.get(dif);
	
	if ( queue == null )
		return;
	
	synchronized ( queue )
	    {
	    if ( queue.size() < MAX_QUEUE_SIZE || psi.compareTo(queue.peek()) > 0 )
	    	{
			System.out.println("Adding = "+psi+" to "+dif+", buffer size = "+(queue.size()+1));
			queue.add(psi);

			fireItemStored(queue, psi);
			
			if ( queue.size() > MAX_QUEUE_SIZE )
				{
				// throw away the least sized item
				@SuppressWarnings("unused")
	            PuzzleStoreItem removed = queue.poll();
				fireItemRemoved(queue, psi);
//						System.out.println("Removed = "+removed);
				}
	    	}
	    }
	}

/**
 * Writes this store to the file system
 * 
 * @throws FileNotFoundException 
 */
void write() throws FileNotFoundException
	{
	File file = new File(filename);
	PrintWriter w = new PrintWriter(file);
	
	try {
		write(w);
		}
	finally
		{
		w.close();
		}
	}

/**
 * Writes this store to the output stream
 * 
 * @param w the writer to write to
 */
private synchronized void write(PrintWriter w)
	{
	pause(true);
	
	for ( PriorityQueue<PuzzleStoreItem> queue : gridStore.values() )
	    {
	    for ( PuzzleStoreItem psi : queue )
	        {
	        w.println(DIFFICULTY+":"+psi.difficulty);
	        w.println(CELLS+":"+psi.nCells);
	        psi.puzzle.write(w);
	        w.println("=====================");
	        }
	    }
	
	pause(false);
	}

/**
 * Reads this file from the input stream
 * 
 * @param r the reader to read from
 * @throws IOException 
 * @throws IllegalGridException 
 */
void read() throws IOException, IllegalGridException
	{
	BufferedReader r = new BufferedReader(new FileReader(filename));

	try 
		{
		read(r);
		}
	finally
		{
		r.close();
		}
	}

/**
 * Reads this file from the input stream
 * 
 * @param r the reader to read from
 * @throws IOException 
 * @throws IllegalGridException 
 */
private synchronized void read(BufferedReader r) throws IOException, IllegalGridException
	{
	pause(true);

	String lineSeparator = System.getProperty("line.separator", "\n");
	StringBuilder sb = new StringBuilder();
	String dif = r.readLine();
	int difficulty = 0;
	int cells = 0;
	
	while ( dif != null )
		{
		if ( dif.startsWith("===") )
			{
			// create this puzzle
			BufferedReader sr = new BufferedReader(new StringReader(sb.toString()));
			Puzzle p = new Puzzle(sr);
			
			PuzzleDifficulty d = p.getDifficulty();
			
			if ( gridStore.containsKey(d) )
				{
				PuzzleStoreItem psi = new PuzzleStoreItem(difficulty, cells, p);
				addToStore(psi);
				}
			
			sb.setLength(0);
			}
		else if ( dif.startsWith(DIFFICULTY) )
			{
			String[] split = dif.split(":");
			difficulty = Integer.parseInt(split[1]);
			}
		else if ( dif.startsWith(CELLS) )
			{
			String[] split = dif.split(":");
			cells = Integer.parseInt(split[1]);
			}
		else
			{
			sb.append(dif);
			sb.append(lineSeparator );
			}
		
		dif = r.readLine();
		}
	
	pause(false);
	}
}
}
