/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JToggleButton;
import javax.swing.SwingUtilities;
import javax.swing.border.Border;
import javax.swing.border.EtchedBorder;
import javax.swing.border.LineBorder;

import uk.co.keang.sudoku.Puzzle.Cell;

/**
 * @author A.G.Docherty
 *
 */
class NumberInputDialog extends JPanel
{
private static final long serialVersionUID = -6830377666453801607L;
private static Dimension panelSize;
private static JDialog dlg;

/**
 * Show this dialog
 * 
 * @param btn
 * @param cell
 * @return the selected numbers or null if cancel was pressed
 */
public static Cell show(SudokuButton btn, Cell cell)
	{
	Frame f = (Frame)SwingUtilities.getAncestorOfClass(Frame.class, btn);	
    dlg = new JDialog(f, "Possible Numbers", true);
    dlg.setLocationRelativeTo(btn);
    
	Border border = btn.getBorder();
	btn.setBorder(new LineBorder(Color.RED, 2));
	NumberInputDialog panel = null;
	
	try {
	    panel = new NumberInputDialog(cell);
	    dlg.add(panel);
	    
	    Dimension size = panelSize;
	    dlg.pack();

	    if ( size != null )
	    	dlg.setSize(size);

	    dlg.setVisible(true);
		}
	finally
		{
		btn.setBorder(border);
		}
	
    return panel.getNumbers();
	}


private static void recordPanelSize()
	{
	panelSize = dlg.getSize();
	}

static Dimension getPanelSize()
	{
	if ( panelSize == null )
		return new Dimension();
	
	return panelSize;
	}

static void setPanelSize(Dimension dim)
	{
	if ( dim.width > 0 && dim.height > 0 )
		panelSize = dim;
	}

private JToggleButton[][] btns;
private final Cell selectedNumbers = new Cell();
protected Cell finalSelection;

/**
 * 
 * @param cell
 */
public NumberInputDialog(Cell nums)
    {
    createButttons();
    initPanel();
    setNumbers(nums);
    }

private void createButttons()
	{
	btns = new JToggleButton[SudokuSolver.SUB_GRID_SIZE][SudokuSolver.SUB_GRID_SIZE];

	ActionListener btnListener = new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			JToggleButton btn = (JToggleButton)e.getSource();
			int num = Integer.parseInt(btn.getText());
			
			if ( !btn.isSelected() )
				selectedNumbers.add(num);
			else
				selectedNumbers.remove(num);
			}
		};

	for ( int y = 0; y < SudokuSolver.SUB_GRID_SIZE; y++ )
	    {
	    for ( int x = 0; x < SudokuSolver.SUB_GRID_SIZE; x++ )
	        {
	        btns[y][x] = new CrossToggleButton(Integer.toString(x+1+(y * SudokuSolver.SUB_GRID_SIZE)));
	        btns[y][x].addActionListener(btnListener);
	        }
	    }
	
	}

private void initPanel()
	{
	setLayout(new BorderLayout(5, 5));
	
	JPanel numbers = new JPanel();
	numbers.setBorder(new EtchedBorder(EtchedBorder.LOWERED));
	numbers.setLayout(new GridLayout(SudokuSolver.SUB_GRID_SIZE, SudokuSolver.SUB_GRID_SIZE, 0, 0));
	
	for ( int y = 0; y < SudokuSolver.SUB_GRID_SIZE; y++ )
	    {
	    for ( int x = 0; x < SudokuSolver.SUB_GRID_SIZE; x++ )
	        {
	        numbers.add(btns[y][x]);
	        }
	    }
	
	add(numbers, BorderLayout.CENTER);
	
	JPanel ctrls = new JPanel();
	ctrls.setLayout(new GridLayout(2, 2, 5, 5));

	JButton clearAll = new JButton("Clear All");
	ctrls.add(clearAll);
	
	JButton selectAll = new JButton("Select All");
	ctrls.add(selectAll);
	
	JButton ok = new JButton("OK");
	ctrls.add(ok);
	
	JButton cancel = new JButton("Cancel");
	ctrls.add(cancel);

	add(ctrls, BorderLayout.SOUTH);
	
	// add listeners
	clearAll.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			clearAllNumbers();
			}
		});
	
	// add listeners
	selectAll.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			selectAllNumbers();
			}
		});
	
	// add listeners
	ok.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			finalSelection = new Cell(selectedNumbers);
			getTopLevelAncestor().setVisible(false);
			}
		});
	
	// add listeners
	cancel.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			finalSelection = null;
			getTopLevelAncestor().setVisible(false);
			}
		});
	

	}

private void setNumbers(Cell nums)
	{
	clearAllNumbers();
	
	for ( int y = 0; y < SudokuSolver.SUB_GRID_SIZE; y++ )
	    {
	    for ( int x = 0; x < SudokuSolver.SUB_GRID_SIZE; x++ )
	        {
	        Integer n = x+1+(y * SudokuSolver.SUB_GRID_SIZE);
	        
	        if ( nums.contains(n) )
	        	{
	        	btns[y][x].setSelected(false);
	        	selectedNumbers.add(n);
	        	}
	        }
	    }
	}

@Override
public void setBounds(int x, int y, int width, int height)
    {
    super.setBounds(x, y, width, height);
    recordPanelSize();
    }

@Override
public void setBounds(Rectangle r)
    {
    super.setBounds(r);
    recordPanelSize();
    }

private Cell getNumbers()
	{
	return finalSelection;
	}

/**
 * Clears all numbers
 */
protected void clearAllNumbers()
    {
    for ( JToggleButton[] row : btns )
        {
        for ( JToggleButton b : row )
            {
        	b.setSelected(true);
            }
        }
    
    selectedNumbers.clear();
    }

/**
 * Selects all numbers
 */
protected void selectAllNumbers()
    {
    for ( JToggleButton[] row : btns )
        {
        for ( JToggleButton b : row )
            {
        	b.setSelected(false);
            }
        }
    
    selectedNumbers.addAll(Puzzle.allPossibleNumbers);
    }

/**
 * A Toggle button that crosses out the selected value
 *  
 * @author A.G.Docherty
 */
private class CrossToggleButton extends JToggleButton
	{
    private static final long serialVersionUID = -2053574797361507500L;

	public CrossToggleButton(String text)
		{
		super(text);
		}
	
	@Override
	public void setBounds(int x, int y, int width, int height)
	    {
	    super.setBounds(x, y, width, height);
	    setFont(getFont().deriveFont(height/2.0f));
	    }
	
	@Override
	public void setBounds(Rectangle r)
	    {
	    super.setBounds(r);
	    setFont(getFont().deriveFont(r.height/2.0f));
	    }

	@Override
	protected void paintComponent(Graphics g)
		{
		super.paintComponent(g);
		
		if ( isSelected() )
			{
			Rectangle bounds = getBounds();
			
			int cross = bounds.height * 5 / 10 / 2;
			
			int cx = (bounds.width/2) - 1;
			int cy = (bounds.height/2) - 1;
			
			g.setColor(Color.RED);
			
			g.drawLine(cx-cross, cy-cross, cx+cross, cy+cross);
			g.drawLine(cx+cross, cy-cross, cx-cross, cy+cross);
			
			if ( cross > 7 )
				{
				g.drawLine(cx-cross+1, cy-cross, cx+cross+1, cy+cross);
				g.drawLine(cx+cross+1, cy-cross, cx-cross+1, cy+cross);
				}
			}
		}
	}

}
