/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku;

import java.awt.Point;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.security.InvalidParameterException;
import java.text.BreakIterator;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.solver.AbstractSolver;
import uk.co.keang.sudoku.solver.AbstractSolver.SolverDifficulty;
import uk.co.keang.sudoku.solver.BruteForceSolver;

/**
 * A Sudoku puzzle
 * 
 * @author A.G.Docherty
 *
 *
 */
public class Puzzle implements Cloneable, Serializable
{
private static final long serialVersionUID = -8679519764942955188L;
public final static Cell allPossibleNumbers;
private static final String[] lblTextXAxis = new String[SudokuSolver.GRID_SIZE];
private static final String[] lblTextYAxis = new String[SudokuSolver.GRID_SIZE];
private Cell[][] hGrid;
private Cell[][] vGrid;
private Cell[][] bGrid;
private Map<Cell, Point> locations;
private PuzzleDifficulty puzzleDifficulty;

public enum PuzzleDifficulty
{
UNKNOWN(0, 0, 0),
TOO_EASY(35, 46, 50),
BEGINNER(50, 36, 50),
EASY(80, 31, 40),
MEDIUM(175, 25, 35),
HARD(300, 20, 35),
EXTREME(Integer.MAX_VALUE, 17, 35);

private int maxVal;
private int minStartingVals;
private int maxStartingVals;

PuzzleDifficulty(int max, int startMin, int startMax)
	{
	maxVal = max;
	minStartingVals = startMin;
	maxStartingVals = startMax;
	}

public int minNumbers() { return minStartingVals; }
public int maxNumbers() { return maxStartingVals; }

int getMinDifficultyValue()
	{
	if ( ordinal() == 0 )
		return 0;
	
	return values()[ordinal()-1].maxVal+1;
	}

static PuzzleDifficulty getDifficulty(int diff)
	{
	for ( PuzzleDifficulty d : values() )
	    {
	    if ( diff <= d.maxVal )
	    	return d;
	    }
	
	throw new InvalidParameterException("No difficulty for "+diff);
	}

@Override
public String toString()
	{
	return setWordCapitalisation(name().replace("_", " "));
	}

/**
 * Set the capitalisation of the text on a word basis.<br>
 * All text to set to lower case and then the first letter of each word is
 * set to upper case. <br>
 * A word is a group of characters following a whitespace character.
 *
 *@param text - the text to capitalise
 *@return String - the capitalised text
 */
public static String setWordCapitalisation(String text)
    {
    StringBuilder sb = new StringBuilder();

    BreakIterator boundary = BreakIterator.getWordInstance();
    boundary.setText(text);

    int start = boundary.first();
    int end = 0;

    while ( (end = boundary.next()) != BreakIterator.DONE )
        {
        char[] c = text.substring(start, end).toLowerCase().toCharArray();

        start = end;

        // if the first char is a letter
        if ( Character.isLetter(c[0]) )
            c[0] = Character.toUpperCase(c[0]);

        sb.append(c);
        }

    return sb.toString();
    }
}

static
	{
	char c = 'A';
	
	for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
		{
		lblTextXAxis[i] = Character.toString((char)(c + i));
		lblTextYAxis[i] = Integer.toString(i+1);
		}
	
	// create a list containing all numbers
	allPossibleNumbers = new Cell();
	
	for ( int i = 1; i <= SudokuSolver.GRID_SIZE; i++ )
	    {
	    allPossibleNumbers.add(i);
	    }
	
	}

public static String[] getXAxisLabels() { return lblTextXAxis; }
public static String[] getYAxisLabels() { return lblTextYAxis; }

/**
 * Constructs an empty puzzle
 * 
 * @param hGrid
 */
public Puzzle()
	{
	locations = new IdentityHashMap<Cell, Point>();
	puzzleDifficulty = PuzzleDifficulty.UNKNOWN;
	
	// create the empty grid
	hGrid = new Cell[SudokuSolver.GRID_SIZE][SudokuSolver.GRID_SIZE];
	vGrid = new Cell[SudokuSolver.GRID_SIZE][SudokuSolver.GRID_SIZE];
	bGrid = new Cell[SudokuSolver.GRID_SIZE][SudokuSolver.GRID_SIZE];
    
    for ( Cell[] row : hGrid )
	    {
	    for ( int i = 0; i < row.length; i++ )
	        {
	        row[i] = new Cell();
	        }
	    }

    createAlternateViews();
    
	clear();
	}

/**
 * Constructs a puzzle from this grid
 * 
 * @param grid
 */
public Puzzle(Cell[][] grid, PuzzleDifficulty dif)
	{
	this();
	
	puzzleDifficulty = dif;
	copyGrid(grid, hGrid);
	}

/**
 * Constructs a puzzle from this input stream
 * @throws IllegalGridException 
 * @throws IOException 
 * 
 */
public Puzzle(BufferedReader r) throws IOException, IllegalGridException
	{
	this();

	read(r);
	}

/**
 * Creates alternate views of the horizontal grid view.
 * These alternate views are for performance issues only - they contain the same set objects
 * as the horizontal grid view.
 */
private void createAlternateViews()
	{
    for ( int y = 0; y < hGrid.length; y++ )
	    {
	    for ( int x = 0; x < hGrid[y].length; x++ )
	        {
	        locations.put(hGrid[y][x], new Point(x, y));
	        vGrid[x][y] = hGrid[y][x];
	        
	        int b = x % SudokuSolver.SUB_GRID_SIZE + ((y % SudokuSolver.SUB_GRID_SIZE) * SudokuSolver.SUB_GRID_SIZE);
	        bGrid[getBlockIndex(x, y)][b] = hGrid[y][x];
	        }
	    }
	}

/**
 * Rotates the grid clockwise 90 degrees
 */
public void rotate()
	{
    for ( int y = 0; y < hGrid.length; y++ )
	    {
	    for ( int x = 0; x < hGrid[y].length; x++ )
	        {
	        hGrid[y][hGrid[y].length-x-1] = vGrid[y][x];
	        }
	    }
	createAlternateViews();
	}

/**
 * Clears this puzzle of all values
 */
public void clear()
	{
	// fill the options list with all possible numbers
	for ( Cell[] rows : hGrid )
	    {
	    for ( Cell sq : rows )
	        {
	        sq.clear();
	        sq.addAll(allPossibleNumbers);
	        }
	    }
	
	puzzleDifficulty = PuzzleDifficulty.UNKNOWN;
	}

/**
 * Makes the sudoku have a unique solution by adding discrete cell values to the starting point
 * This can only be successfully called on puzzles that have a solution.
 * 
 * @return The number of values added
 */
public int makeUnique()
	{
	int count = 0;
	
	try
	    {
		Puzzle p = clone();
	    Point loc = findVariationPoint(p);

	    while ( loc != null )
	    	{
		    hGrid[loc.y][loc.x].clear();
		    hGrid[loc.y][loc.x].addAll(p.getGrid()[loc.y][loc.x]);
		    count ++;
		    
			p = clone();
		    loc = findVariationPoint(p);
	    	}
	    }
    catch ( IllegalGridException exc )
	    {
	    return 0;
	    }
    
    return count;
	}

/**
 * Checks to see if the sudoku is unique by using the brute force solver several times
 * and comparing the complete grids 
 * 
 * @return true if this has a unique solution. False if the solution is not unique or not possible
 */
public boolean isUnique()
	{
	try
	    {
	    return findVariationPoint(this) == null;
	    }
    catch ( IllegalGridException exc )
	    {
	    // Puzzle not solvable so the solution can't be unique
	    return false;
	    }
	}
	
/**
 * Finds the point at which variation can occur in solutions
 * 
 * @param p the puzzle to solve - this will be modified
 * @return the location that variation can occur or null if no variation occurs. 
 * @throws IllegalGridException If the puzzle is not solvable 
 */
private Point findVariationPoint(Puzzle p) throws IllegalGridException
	{
	Puzzle p1 = p.clone();
	AbstractSolver solver = new BruteForceSolver(p1, new Random(System.currentTimeMillis()));
	solver.findSolutions();
    
    for ( int i = 0; i < 2; i++ )
	    {
	    // solve it from the other direction
		Puzzle p2 = p.clone();
		p2.rotate();
		p2.rotate();
		
		AbstractSolver solver1 = new BruteForceSolver(p2, new Random(System.currentTimeMillis()));
		solver1.findSolutions();

		p2.rotate();
		p2.rotate();
		
        // find the variation point
    	for ( int y = 0; y < hGrid.length; y++ )
    	    {
    	    for ( int x = 0; x < hGrid[y].length; x++ )
    	        {
    	        if ( !p1.hGrid[y][x].equals(p2.hGrid[y][x]) )
    	        	return new Point(x, y);
    	        }
    	    }
	    }
	
	return null;
	}


/**
 * Checks to see if the sudoku is complete
 * 
 * @return true if all cells have been resolved and all of the values are legal.
 */
public boolean isComplete()
	{
	for ( int y = 0; y <SudokuSolver.GRID_SIZE; y++ )
	    {
	    for ( int x = 0; x <SudokuSolver.GRID_SIZE; x++ )
	        {
	        if ( hGrid[y][x].size() != 1 )
	        	return false;
	        }
	    }
	
	return true;
	}

/**
 * Checks all resolved values to check they are legal
 * 
 * @return true if all resolved values are legal
 */
public boolean isLegal()
	{
	for ( int y = 0; y <SudokuSolver.GRID_SIZE; y++ )
	    {
	    for ( int x = 0; x <SudokuSolver.GRID_SIZE; x++ )
	        {
	        if ( hGrid[y][x].size() == 0 )
	        	return false;
	        
	        if ( hGrid[y][x].size() == 1 )
	        	{
	        	// this square has a resolved value check it is legal
	        	int val = hGrid[y][x].firstValue();
	        	int count = 0;
	        	
	        	// count how many times it is in the row
	        	for ( Cell sq : hGrid[y] )
	                {
	                if ( sq.size() == 1 && sq.contains(val) )
	                	count++;
	                }
	        	
	        	// count how many times it is in the column
	        	for ( int i = 0; i <SudokuSolver.GRID_SIZE; i++ )
	                {
	                Cell sq = hGrid[i][x];

	                if ( sq.size() == 1 && sq.contains(val) )
	                	count++;
	                }

	        	// count how many times it is in the block
	        	int x1 = (x/SudokuSolver.SUB_GRID_SIZE) * SudokuSolver.SUB_GRID_SIZE;
	        	int y1 = (y/SudokuSolver.SUB_GRID_SIZE) * SudokuSolver.SUB_GRID_SIZE;
	        	
	        	for ( int i = 0; i < SudokuSolver.SUB_GRID_SIZE; i++ )
	                {
	                for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
	                    {
		                Cell sq = hGrid[y1+i][x1+j];

		                if ( sq.size() == 1 && sq.contains(val) )
		                	count++;
	                    }
	                } 
	        	
	        	if ( count != 3 )
	        		return false;
	        	}
	        }
	    }
	
	return true;
	}

/**
 * Gets the possible values for this point 
 * @param p
 * @return
 */
public Cell getPossibleValues(Point p)
	{
	// if this point only has 1 value return it now
	if ( hGrid[p.y][p.x].size() == 1 )
		return new Cell(hGrid[p.y][p.x]);

	// work out which values it can be
	Cell allVals = new Cell(allPossibleNumbers);
	Cell knownVals = getKnownValues(p);
	
	allVals.removeAll(knownVals);
	return allVals;
	}

/**
 * Gets the known values from the row, column or block for this point 
 * @param p
 * @return
 */
public Cell getKnownValues(Point p)
	{
	Cell vals = new Cell(); 
	
	// check this row for duplicates
	Cell[] row = hGrid[p.y];
	getKnownValues(vals, row);
	
	// check this column for duplicates
	Cell[] col = getColumn(p.x);
	getKnownValues(vals, col);
	
	// check this block for duplicates
	Cell[] block = getBlock(p.x, p.y);
	getKnownValues(vals, block);

	return vals;
	}

/**
 * Gets all known values in the group 
 * 
 * @param grp
 * @param index
 * @return
 */
private void getKnownValues(Cell vals, Cell[] grp)
	{
	for ( Cell set : grp )
	    {
	    if ( set.size() == 1 )
	    	{
	    	vals.addAll(set);
	    	}
	    }
	}


/**
 * Gets the options column
 * 
 * @param i the 0 based column number
 * @return
 */
public Cell[] getColumn(int i)
	{
	return vGrid[i];
	}

/**
 * Gets the options block
 * 
 * @param x the 0 based x coordinate
 * @param y the 0 based y coordinate
 * @return
 */
public int getBlockIndex(int x, int y)
	{
	int x1 = x/SudokuSolver.SUB_GRID_SIZE;
	int y1 = y/SudokuSolver.SUB_GRID_SIZE;

	return (y1*SudokuSolver.SUB_GRID_SIZE)+x1;
	}


/**
 * Gets the options block
 * 
 * @param x the 0 based x coordinate
 * @param y the 0 based y coordinate
 * @return
 */
public Cell[] getBlock(int x, int y)
	{
	return getBlock(getBlockIndex(x, y));
	}

/**
 * Gets the options block
 * 
 * @param i the 0 based block number
 * @return
 */
public Cell[] getBlock(int i)
	{
	return bGrid[i];
	}

/**
 * Gets the zero based x, y coordinates of this set in the grid
 * 
 * @param set the set to find
 * @return the x,y coordinates
 */
public Point getLocation(Cell set)
	{
	Point loc = locations.get(set);
	
	if ( loc == null )
		throw new IllegalArgumentException("The set does not belong to the grid");
	
	return loc;
	}

/**
 * Gets these locations as string of textual coordinates with letters for the x axis and 
 * numbers for the y axis
 * 
 * @param loc
 * @return
 */
public static String getLocationsAsText(List<Point> loc)
	{
	String text;
	if ( loc.size() == 1 )
		{
		Point p = loc.iterator().next();
		text = lblTextXAxis[p.x] + lblTextYAxis[p.y];
		}
	else
		{
		StringBuffer sb = new StringBuffer();
		for ( int i = 0, j = loc.size(); i < j; i++ )
			{
			Point p = loc.get(i);
			
			if ( i == j-1 )
				sb.append(" and ");
			else if ( i > 0 )
				sb.append(", ");

			sb.append(lblTextXAxis[p.x]);
			sb.append(lblTextYAxis[p.y]);
			}

		text = sb.toString();
		}
	
	return text;
	}

/**
 * Gets these numbers as a formatted text string
 * 
 * @param cell
 * @return
 */
public static String getNumbersAsText(Cell cell)
	{
	return getNumbersAsText(cell.nums);
	}

/**
 * Gets these numbers as a formatted text string
 * 
 * @param removed
 * @return
 */
public static String getNumbersAsText(Collection<Integer> removed)
	{
	String nums;
	if ( removed.size() == 1 )
		{
		nums = removed.iterator().next().toString();
		}
	else
		{
		List<Integer> list = new ArrayList<Integer>(removed);
		Collections.sort(list);
		StringBuffer sb = new StringBuffer();
		for ( int i = 0, j = list.size(); i < j; i++ )
			{
			if ( i == j-1 )
				sb.append(" or ");
			else if ( i > 0 )
				sb.append(", ");

			sb.append(list.get(i));
			}

		nums = sb.toString();
		}
	return nums;
	}

/**
 * Gets the frequency of possible values. Known values are ignored.
 * 
 * @param sets the line of sets to get the frequencies from
 * @return an array where each index represents the number - 1. The value is the frequency of
 * 			occurrence of possible values and where - 1 signifies this number is known. 
 */
public static int[] getFrequencyOfPossibilities(Cell[] sets)
	{
	int[] occur = new int[SudokuSolver.GRID_SIZE];

	// find how many times each number occurs in this line
	// resolved numbers are not counted
	for ( Cell s : sets )
		{
		if ( s.size() == 1 )
			{
			// this is a known number
			int val = s.firstValue();
			occur[val-1] = -1;
			}
		else
			{
			// increment the counters for each number
			for ( Integer val : s.nums )
				{
				val--;

				if ( occur[val] != -1 )
					occur[val]++;
				}
			}
		}

	return occur;
	}

@Override
public Puzzle clone() 
    {
    Puzzle p = null;
    try
	    {
	    p = (Puzzle)super.clone();
	    }
    catch ( CloneNotSupportedException exc )
	    {
	    // Clone is supported so this won't ever happen
	    exc.printStackTrace();
	    }
    
    p.hGrid = copyGrid(hGrid);
    p.puzzleDifficulty = puzzleDifficulty;
    
    p.locations = new IdentityHashMap<Cell, Point>();
	p.vGrid = new Cell[SudokuSolver.GRID_SIZE][SudokuSolver.GRID_SIZE];
	p.bGrid = new Cell[SudokuSolver.GRID_SIZE][SudokuSolver.GRID_SIZE];

    p.createAlternateViews();
    return p;
    }

/**
 * creates an copy of the grid
 * 
 * @param grid
 * @return the copy
 */
static Cell[][] copyGrid(Cell[][] grid)
	{
    Cell[][] copy = new Cell[SudokuSolver.GRID_SIZE][SudokuSolver.GRID_SIZE];
	
	for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
	    {
	    for ( int j = 0; j < SudokuSolver.GRID_SIZE; j++ )
	        {
	        copy[i][j] = new Cell(grid[i][j]);
	        }
	    }

	return copy;
	}

/**
 * Copies this grid to another grid
 * 
 * @param grid
 * @return the destination
 */
static Cell[][] copyGrid(Cell[][] src, Cell[][] dest)
	{
	for ( int y = 0; y < SudokuSolver.GRID_SIZE; y++ )
	    {
	    for ( int x = 0; x < SudokuSolver.GRID_SIZE; x++ )
	        {
	        dest[y][x].copy(src[y][x]);
	        }
	    }
	
	return dest;
	}

/**
 * Gets this puzzle grid
 * 
 * @return
 */
public Cell[][] getGrid()
	{
	return hGrid;
	}

/**
 * Sets the value at square denoted by the sequential (1 based) location from top left to bottom right.
 * Cell 1 is at the top left and the cells are contiguously numbered from 
 * left to right. At the end of the row the next cell is the leftmost cell on the
 * next row down. 
 * 
 * @param i the 1 based cell number
 * @param num the number. If the number is not legal ie < 1 or > Sudoku.GRID_SIZE the call is set to all possible values 
 * @return the original value
 */
public Cell set(int i, int num)
	{
	int y = (i-1) / SudokuSolver.GRID_SIZE;
	int x = i % SudokuSolver.GRID_SIZE;
	
	if ( x == 0 )
		x = SudokuSolver.GRID_SIZE;
	
	x--;
	
	return set(x, y, num);
	}

/**
 * Sets the value at square denoted by the x, y (0 based) location
 * 
 * @param x the 0 based x value
 * @param y the 0 based y value
 * @param num the number
 * @return the original value
 */
public Cell set(int x, int y, int num)
	{
	if ( num > 0 && num <= SudokuSolver.GRID_SIZE )
		{
		// use this legal value 
		Cell val = new Cell();
		val.add(num);
		return set(x, y, val);
		}
		
	return set(x, y, allPossibleNumbers);
	}

/**
 * Sets the values at square denoted by the x, y (0 based) location.
 * If the set of numbers is empty then all possible numbers will be used
 * 
 * @param x the 0 based x value
 * @param y the 0 based y value
 * @param num the numbers
 * @return the original value
 */
public Cell set(int x, int y, Cell nums)
	{
	Cell old = new Cell(hGrid[y][x]);
	
	hGrid[y][x].clear();
	
	if ( nums == null || nums.isEmpty() )
		hGrid[y][x].addAll(allPossibleNumbers);
	else
		hGrid[y][x].addAll(nums);
	
	return old;
	}

/**
 * Sets the values for each cell of the sudoku grid. 
 * A cell with no values will be assigned all possible values
 * 
 * @param a 2D array of all possible values for each cell
 */
public void setGrid(Cell[][] grid)
	{
	for ( int y = 0; y < grid.length; y++ )
	    {
	    for ( int x = 0; x < grid[y].length; x++ )
	        {
	        set(x, y, grid[y][x]);
	        }
	    }
	}

@Override
public String toString()
    {
	StringBuffer sb = new StringBuffer();
	for ( int i = 0; i < SudokuSolver.GRID_SIZE*2+4; i++ )
	    {
	    sb.append("-");
	    }
	
	String line = sb.toString();

	sb.setLength(0);
	
    for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
	    {
	    Cell[] x = hGrid[i];
	    
        if ( i % SudokuSolver.SUB_GRID_SIZE == 0 )
        	{
        	sb.append(line);
        	sb.append("\n");
        	}
        
	    for ( int j = 0; j < x.length; j++ )
	        {
	        if ( j % SudokuSolver.SUB_GRID_SIZE == 0 )
	    		sb.append("|");

	        if ( x[j].size() == 1 )
	    		sb.append(x[j].firstValue());
	        else
	    		sb.append(".");

    		sb.append(" ");
	        }
	    
		sb.append("|\n");
	    }

	sb.append(line);
	sb.append("\n");

	//
	
    for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
	    {
	    Cell[] x = vGrid[i];
	    
        if ( i % SudokuSolver.SUB_GRID_SIZE == 0 )
        	{
        	sb.append(line);
        	sb.append("\n");
        	}
        
	    for ( int j = 0; j < x.length; j++ )
	        {
	        if ( j % SudokuSolver.SUB_GRID_SIZE == 0 )
	    		sb.append("|");

	        if ( x[j].size() == 1 )
	    		sb.append(x[j].firstValue());
	        else
	    		sb.append(".");

    		sb.append(" ");
	        }
	    
		sb.append("|\n");
	    }

	sb.append(line);
	sb.append("\n");

	
	return sb.toString();
    }

@Override
public boolean equals(Object obj)
    {
    if ( obj == null || obj.getClass() != this.getClass() )
    	return false;
    
    Puzzle p = (Puzzle)obj;
    
    if ( hGrid.length != p.hGrid.length )
    	return false;
    
	for ( int y = 0; y < hGrid.length; y++ )
	    {
	    for ( int x = 0; x < hGrid[y].length; x++ )
	        {
	        if ( !hGrid[y][x].equals(p.hGrid[y][x]) )
	        	return false;
	        }
	    }
	
	return true;
    }

@Override
public int hashCode()
    {
    int val = 1;
    
	for ( int y = 0; y < hGrid.length; y++ )
	    {
	    for ( int x = 0; x < hGrid[y].length; x++ )
	        {
	        val *= hGrid[y][x].hashCode();
	        }
	    }
	
	return val;
    }

/**
 * Counts the number of known values in this puzzle
 * 
 * @return
 */
public int countKnownValues()
	{
	int count = 0;
	for ( int y = 0; y < hGrid.length; y++ )
	    {
	    for ( int x = 0; x < hGrid[y].length; x++ )
	        {
	        if ( hGrid[y][x].size() == 1 )
	        	count++;
	        }
	    }
	
	return count;
	}

/**
 * Writes this file to the output stream
 * 
 * @param w the writer to write to
 */
public void write(PrintWriter w)
	{
	// write the difficulty
	w.print(puzzleDifficulty.name());
	w.println();
	
	// write the grid
	StringBuilder sb = new StringBuilder();
    for ( int y = 0; y < hGrid.length; y++ )
	    {
	    Cell[] line = hGrid[y];
	    
	    // for each cell in this row
	    for ( int x = 0; x < line.length; x++ )
	        {
	        // add each number in this cell
        	for ( Integer n : hGrid[y][x] )
                {
            	sb.append(n);
            	sb.append(",");
                }

        	sb.setLength(sb.length()-1);
        	
        	if ( hGrid[y][x].userEntered )
        		sb.append(";U");
        	
        	sb.append(":");
	        }
	    
	    sb.setLength(sb.length()-1);
	    
		w.print(sb.toString());
		w.println();
		
		sb.setLength(0);
	    }
	
	}

/**
 * Reads this file from the input stream
 * 
 * @param r the reader to read from
 * @throws IOException 
 * @throws IllegalGridException 
 */
public void read(BufferedReader r) throws IOException, IllegalGridException
	{
    String dif = r.readLine();
    puzzleDifficulty = Enum.valueOf(PuzzleDifficulty.class, dif.trim().toUpperCase());
    
	for ( int y = 0; y < SudokuSolver.GRID_SIZE; y++ )
	    {
	    // read each row
	    String line = r.readLine();
	    String[] cells = line.split(":");
	    
	    if ( cells.length != SudokuSolver.GRID_SIZE )
	    	throw new IllegalGridException(null);
	    
	    // for each cell in this row
	    for ( int x = 0; x < cells.length; x++ )
	        {
	        hGrid[y][x].clear();

		    String[] user = cells[x].split(";");
		    
		    String cNums;
			if ( user.length > 1 )
		    	{
		    	if ( user[1].contains("U") )
		    		hGrid[y][x].userEntered = true;
		    	
		    	cNums = user[0];
		    	}
		    else
		    	cNums = cells[x];
		    
		    String[] nums = cNums.split(",");

		    for ( String n : nums )
	            {
	        	int i = Integer.parseInt(n);
	        	hGrid[y][x].add(i);
	            }
	        }
	    }
	
	}

/**
 * Sets an level indicating how difficult it is to solve this puzzle
 * 
 * @param solver the solver difficulty
 * @param factor the difficulty factor
 */
public void setDifficulty(SolverDifficulty solver, int factor)
	{
	PuzzleDifficulty sDiff = PuzzleDifficulty.valueOf(solver.name());
	PuzzleDifficulty fDiff = PuzzleDifficulty.getDifficulty(factor);
	
	// if the solver is harder than the factor then use the solver difficulty value
	if ( sDiff.ordinal() >= fDiff.ordinal() )
		puzzleDifficulty = sDiff;
	else
		puzzleDifficulty = fDiff;
	}

/**
 * Sets an level indicating how difficult it is to solve this puzzle
 * 
 * @param dif the difficulty
 */
public void setDifficulty(PuzzleDifficulty dif)
	{
	puzzleDifficulty = dif;
	}

/**
 * Gets an indication of how difficult it is to solve this puzzle
 * 
 * @return the difficulty
 */
public PuzzleDifficulty getDifficulty()
	{
	return puzzleDifficulty;
	}

//=====================================================================================================
public static class Cell implements Iterable<Integer>
{
Set<Integer> nums = new HashSet<Integer>();
boolean userEntered = false;

public Cell()
	{
	// create an empty cell
	}

public Cell(Collection<Integer> c)
	{
	nums.addAll(c);
	}

public Cell(Cell cell)
	{
	addAll(cell);
	userEntered = cell.userEntered;
	}

public void clear()
	{
	nums.clear();
	userEntered = false;
	}

public int size()
	{
	return nums.size();
	}

public boolean isEmpty()
	{
	return nums.isEmpty();
	}

public boolean contains(Integer val)
	{
	return nums.contains(val);
	}

public boolean containsAll(Cell cell)
	{
	return nums.containsAll(cell.nums);
	}

public void add(Integer num)
	{
	nums.add(num);
	}

public void addAll(Cell set)
	{
	nums.addAll(set.nums);
	}

public void copy(Cell set)
	{
	nums.clear();
	nums.addAll(set.nums);
	userEntered = set.userEntered;
	}

public void remove(Integer num)
	{
	nums.remove(num);
	}

public void removeAll(Cell knownVals)
	{
	nums.removeAll(knownVals.nums);
	}

public Integer firstValue()
	{
	return nums.iterator().next();
	}

@Override
public Iterator<Integer> iterator()
	{
	return nums.iterator();
	}

public void retainAll(Cell retainVals)
	{
	nums.retainAll(retainVals.nums);
	}

public Collection<Integer> getCollection()
	{
	return nums;
	}

@Override
public String toString()
    {
    return nums.toString();
    }

@Override
public boolean equals(Object obj)
    {
    if ( !(obj instanceof Cell) )
    	return false;
    
    return nums.equals(((Cell)obj).nums);
    }

@Override
public int hashCode()
    {
    return nums.hashCode();
    }
}
}


