/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;

import javax.swing.CellRendererPane;
import javax.swing.JButton;
import javax.swing.JLabel;

import uk.co.keang.sudoku.Puzzle.Cell;

/**
 * @author A.G.Docherty
 *
 */
public class SudokuButton extends JButton
{
private static final long serialVersionUID = -6731274518042923584L;
private final CellRendererPane cellRendererPane = new CellRendererPane();
private final Cell possibleNums = new Cell();
private final Cell removeNums = new Cell();
private final JLabel stamp = new JLabel("", JLabel.CENTER);
private Font bigFont;
private Font littleFont;
private final int xLocation;
private final int yLocation;
private BtnState btnState = BtnState.NORMAL;
private boolean showPossibleNumbers;

public enum BtnState
{
SELECTED(Color.GREEN, Color.RED),
REFERENCE(Color.WHITE, Color.GREEN),
NORMAL(Color.LIGHT_GRAY, Color.BLACK),
ERROR(Color.RED, Color.BLACK);

private Color bkgColour;
private Color frgColour;

private BtnState(Color bk, Color fore)
	{
	bkgColour = bk;
	frgColour = fore;
	}

Color getBkgColor() { return bkgColour; }
Color getFrgColor() { return frgColour; }
}

public SudokuButton(int x, int y)
	{
	xLocation = x;
	yLocation = y;
	}


/**
 * Sets the resolved value for this cell
 * 
 * @param resolvedValue
 */
public void setResolved(int resolvedValue)
	{
	possibleNums.clear();
	possibleNums.add(resolvedValue);
	removeNums.clear();
	}

/**
 * Sets the current possible numbers for this cell
 * 
 * @param set
 */
public void setPossibilites(Cell set)
	{
	possibleNums.clear();
	possibleNums.addAll(set);
	possibleNums.userEntered = set.userEntered;
	removeNums.clear();
	}

/**
 * Gets the current possible numbers for this cell
 * 
 * @return the possible numbers
 */
public Cell getPossibilites()
	{
	return new Cell(possibleNums);
	}

/**
 * Marks these numbers for removal from this cell 
 * @param set the numbers to show as being removed or null to clear the previous values
 */
public void setRemovePossibilites(Cell set)
	{
	removeNums.clear();
	
	if ( set != null )
		removeNums.addAll(set);
	}

/**
 * Sets the current display state of this button
 * @param state
 */
public void setDisplayState(BtnState state)
	{
	btnState = state;
	}

/**
 * Sets the show possible numbers state
 * 
 * @param show true to show all possible numbers, false to just show those entered by the user
 */
void showPossibleNumbers(boolean show)
	{
	showPossibleNumbers = show;
	}

/**
 * Gets the 0 based grid location of this button
 * 
 * @return the grid location
 */
Point getGridLocation()
	{
	return new Point(xLocation, yLocation);
	}

@Override
public void setBounds(int x, int y, int w, int h)
	{
	createFonts(h);
	super.setBounds(x, y, w, h);
	}

private void createFonts(int height)
	{
	bigFont = getFont().deriveFont(height*0.75f);
	littleFont = getFont().deriveFont(((height-2.0f)/3)-2.0f);
	}

@Override
public void paintComponent(Graphics g)
	{
//	setBackground(btnState.getBkgColor());
//	super.paintComponent(g);
	g.setColor(btnState.getBkgColor());
	g.fill3DRect(0, 0, getWidth()-1, getHeight()-1, true);
	stamp.setOpaque(false);
	
	int x, y, w, h;
	
	if ( possibleNums.size() == 1 )
		{
		// paint a large single number
		stamp.setFont(bigFont);
		stamp.setForeground(BtnState.NORMAL.getFrgColor());
		stamp.setText(Integer.toString(possibleNums.iterator().next()));
		
		x = 0;
		y = 0;
		
		Dimension d = getSize();
		w = d.width;
		h = d.height;
		cellRendererPane.paintComponent(g, stamp, this, x-1, y-1, w, h);
		}
	else 
		{
		stamp.setFont(littleFont);
		stamp.setForeground(Color.BLACK);
		
		Dimension d = getSize();
		w = (d.width-4)/3;
		h = (d.height-4)/3;
		x = 0;
		y = 0;

		// for each number add it to the next space on the grid
		for ( Integer num : possibleNums )
	        {
    		stamp.setOpaque(false);
    		
	        // move stamp to the correct location
	        stamp.setText(Integer.toString(num));
	        
	        // if the number is being removed colour it red
	        if ( removeNums.contains(num) )
	        	{
	        	if ( btnState == BtnState.REFERENCE )
	        		{
	        		stamp.setOpaque(true);
		    		stamp.setForeground(Color.WHITE);
		    		stamp.setBackground(btnState.getFrgColor());
	        		}
	        	else
	        		{
		    		stamp.setForeground(btnState.getFrgColor());
	        		}

	        	}
	        else
	        	{
	    		stamp.setForeground(BtnState.NORMAL.getFrgColor());
	        	}
	        
	        boolean draw = showPossibleNumbers || possibleNums.userEntered || btnState != BtnState.NORMAL;
	        if ( draw )
	        	cellRendererPane.paintComponent(g, stamp, this, x+2, y+2, w, h);

			x += w;
			
			if ( x >= w*SudokuSolver.SUB_GRID_SIZE )
				{
				x = 0;
				y += h;
				}
	        }
		
		}
	
	}

}
