/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.concurrent.ExecutionException;

import javax.swing.JApplet;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSeparator;
import javax.swing.ProgressMonitor;
import javax.swing.SwingWorker;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.border.EmptyBorder;

import uk.co.keang.sudoku.Generator.PuzzleStoreItem;
import uk.co.keang.sudoku.Puzzle.PuzzleDifficulty;
import uk.co.keang.sudoku.event.IllegalGridException;

/**
 * A GUI interface for the Sudoku solver
 * @author A.G.Docherty
 *
 */
public class SudokuGUI extends JPanel implements StoreListener
{
private static final long serialVersionUID = 6876807083281886901L;
private static final int MAX_PROGRESS = 100;
private static final int MAX_SNAPSHOTS = 10;
private static Dimension screenSize;

private final SudokuSolver sudokuSolver;
private final Generator generator;
private SudokuGrid gridPanel;
private JMenu generateMenu;
private boolean autoGenerate = true;
private File currentFile;
private boolean autoSolve = true;
private SolutionPanel solutionsPanel;
private boolean showPossibleNums = true;
private final TopLevelWindow frameType;
private final List<Snapshot> snapshots = new ArrayList<Snapshot>();
private final Map<Puzzle.PuzzleDifficulty, JMenuItem> menuMap = new EnumMap<Puzzle.PuzzleDifficulty, JMenuItem>(Puzzle.PuzzleDifficulty.class);
private boolean autoSave = true;
private final SequenceBuffer seqBuff;
protected boolean restoreOnStartUp;
private JCheckBoxMenuItem autoSolveMenu;
private JCheckBoxMenuItem showNumbersMenu;
private JCheckBoxMenuItem restoreOnStartUpMenu;
private JCheckBoxMenuItem autoSaveMenu;
private JCheckBoxMenuItem autoGenerateMenu;

static
{
screenSize = Toolkit.getDefaultToolkit().getScreenSize();
}

private enum TopLevelWindow
{
FRAME {
		@Override
		void setMenuBar(JMenuBar mb)
			{
			((JFrame)top).setJMenuBar(mb);
			}

		@Override
        void setTitle(String title)
	        {
	        ((JFrame)top).setTitle(title);
	        }
	},
DIALOG {
		@Override
		void setMenuBar(JMenuBar mb)
			{
			((JDialog)top).setJMenuBar(mb);
			}

		@Override
        void setTitle(String title)
	        {
	        ((JDialog)top).setTitle(title);
	        }
	},
APPLET {
		@Override
		void setMenuBar(JMenuBar mb)
			{
			((JApplet)top).setJMenuBar(mb);
			}

		@Override
        void setTitle(String title)
	        {
	        // do nothing as there is no title
	        }

		@Override
        void setWindowClosedListener(final SudokuGUI gui)
	        {
	        // do nothing as there is no window to listen for
	        }

		@Override
        void setWindowOpenedListener(final ApplicationMemento mem)
	        {
	        // do nothing as there is no window to listen for
	        }
	};

private static Container top;

static TopLevelWindow getEnum(Container c)
	{
	top = c;
	if ( c instanceof JFrame )
		{
		return FRAME;
		}
	else if ( c instanceof JDialog )
		{
		return DIALOG;
		}
	else if ( c instanceof JApplet )
		{
		return APPLET;
		}
	else
		{
		throw new IllegalStateException("The only valid top level windows are JFrame, JApplet and JDialog");
		}
	}

abstract void setMenuBar(JMenuBar mb);

abstract void setTitle(String title);

void setWindowClosedListener(final SudokuGUI gui)
	{
	((Window)top).addWindowListener(new WindowAdapter()
		{
		@Override
		public void windowClosed(WindowEvent e)
		    {
		    gui.windowClosed();
		    }
		});
	}

void setWindowOpenedListener(final ApplicationMemento mem)
	{
	((Window)top).addWindowListener(new WindowAdapter()
		{
		@Override
		public void windowOpened(WindowEvent e)
		    {
		    mem.restoreValues();
		    
		    ((Window)top).removeWindowListener(this);
		    }
		});
	}

Dimension getSize()
	{
	return top.getSize();
	}

Point getLocation()
	{
	return top.getLocation();
	}

void setSize(Dimension dimension)
	{
	top.setSize(dimension);
	}

void setLocation(Point point)
	{
	top.setLocation(point);
	}
}
/**
 * Displays this panel in a centrally positioned JFrame 
 *
 * @throws DataStoreException
 */
public static void show(String[] puzzle)
    {
    JFrame dlg = new JFrame();
    dlg.setTitle("Sudoku");
    dlg.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    SudokuGUI panel = new SudokuGUI(dlg, puzzle);
    
    dlg.add(panel);

    dlg.pack();
    Dimension d = dlg.getSize();
//    dlg.setSize(Math.max(d.width, 898), Math.max(d.height+30, 492));
    dlg.setSize(896, 463);

    d = dlg.getSize();
    dlg.setLocation((screenSize.width-d.width)/2, (screenSize.height-d.height)/2);

    dlg.setResizable(true);
    dlg.setVisible(true);
    }

/**
 * Creates the Sudoku GUI
 * 
 * @param owner
 * @param args
 */
public SudokuGUI(Container owner, String[] args)
    {
    sudokuSolver = new SudokuSolver(args);
    seqBuff = new SequenceBuffer();
    sudokuSolver.setSequenceBuffer(seqBuff);

    frameType = TopLevelWindow.getEnum(owner);
    initMenu(frameType);
    initPanel();
    
    frameType.setWindowClosedListener(this);
    
    // load the saved application settings
    ApplicationMemento mem = new ApplicationMemento();
    try
	    {
	    mem.load();
	    frameType.setWindowOpenedListener(mem);
	    }
    catch ( FileNotFoundException exc )
	    {
	    // Do nothing - there's no file to load
	    }
    catch ( IOException exc1 )
	    {
	    exc1.printStackTrace();
	    }
    catch ( IllegalGridException exc1 )
	    {
	    exc1.printStackTrace();
	    }
    catch ( Exception exc1 )
	    {
	    exc1.printStackTrace();
	    }
    
    generator = new Generator();
    generator.addStoreListener(this);
    try
	    {
	    // if we were saving the puzzle store then load it now
	    if ( autoSave )
	    	generator.load();
	    }
    catch ( FileNotFoundException exc )
	    {
	    // Do nothing - there's no file to load
	    }
    catch ( IOException exc )
	    {
	    exc.printStackTrace();
	    }
    catch ( IllegalGridException exc )
	    {
	    exc.printStackTrace();
	    }
    
	setAutoGeneratorState(autoGenerate);
    }

/**
 * Creates the menu components
 * 
 * @param frameType2 the top level container which will display the menu bar
 */
private void initMenu(TopLevelWindow win)
	{
	JMenuBar mb = new JMenuBar();
	
	win.setMenuBar(mb);

	// Create the File Menu
	mb.add(createFileMenu());
	
	// Create the Generate Menu
	mb.add(createGenerateMenu());
	
	// Create the Snapshot Menu
	mb.add(createSnapshotMenu());

	// Create the Options Menu
	mb.add(createOptionsMenu());

	// Create the About Menu
	mb.add(createAboutMenu());
	}

/**
 * Creates the file menu
 * 
 * @return the menu 
 */
private JMenu createFileMenu()
	{
	JMenu fileMenu = new JMenu("File");
	
	JMenuItem mItem = new JMenuItem("Load");
	fileMenu.add(mItem);
	
	mItem.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			try
	            {
	            File file = Storage.selectFile(SudokuGUI.this, currentFile, "Load");
	            
	            if ( file != null )
	            	{
	            	currentFile = file;
	            	frameType.setTitle("Sudoku - "+currentFile.getName());

	            	Puzzle p = Storage.load(currentFile);
	            	gridPanel.setPuzzle(p);
	            	}
	            }
            catch ( IOException exc )
	            {
	            exc.printStackTrace();
	            JOptionPane.showMessageDialog(SudokuGUI.this, "File failed to load due to:\n"+exc.getMessage(), "Load failed", JOptionPane.ERROR_MESSAGE);
	            }
            catch ( IllegalGridException exc )
	            {
	            exc.printStackTrace();
	            JOptionPane.showMessageDialog(SudokuGUI.this, "File failed to load as the file is not formatted correctly", "Load failed", JOptionPane.ERROR_MESSAGE);
	            }
			}
		});
	
	fileMenu.add(new JSeparator());
	
	mItem = new JMenuItem("Save");
	fileMenu.add(mItem);
	
	mItem.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// SAVE 
			try
	            {
	            Puzzle p = new Puzzle(gridPanel.getDisplayedValues(), sudokuSolver.getPuzzle().getDifficulty());
	            
	            if ( currentFile == null || currentFile.isDirectory() )
	            	currentFile = Storage.save(p, SudokuGUI.this, currentFile);
	            else
	            	Storage.save(p, currentFile);

	            if ( currentFile != null )
	            	frameType.setTitle("Sudoku - "+currentFile.getName());
	            }
            catch ( IOException exc )
	            {
	            exc.printStackTrace();
	            JOptionPane.showMessageDialog(SudokuGUI.this, "File failed to save due to:\n"+exc.getMessage(), "Save failed", JOptionPane.ERROR_MESSAGE);
	            }
			}
		});
	
	mItem = new JMenuItem("Save As...");
	fileMenu.add(mItem);
	
	mItem.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// SAVE AS
            Puzzle p = new Puzzle(gridPanel.getDisplayedValues(), sudokuSolver.getPuzzle().getDifficulty());

            try
	            {
	            File file = Storage.save(p, SudokuGUI.this, currentFile);
	            
	            if ( file != null )
	            	{
	            	currentFile = file;
	            	frameType.setTitle("Sudoku - "+currentFile.getName());
	            	}
	            }
            catch ( IOException exc )
	            {
	            exc.printStackTrace();
	            JOptionPane.showMessageDialog(SudokuGUI.this, "File failed to save due to:\n"+exc.getMessage(), "Save failed", JOptionPane.ERROR_MESSAGE);
	            }
			}
		});
	
	return fileMenu;
	}

/**
 * Creates the Generate menu
 * 
 * @return the menu 
 */
private JMenu createGenerateMenu()
	{
	generateMenu = new JMenu("Puzzles");
	
	autoGenerateMenu = new JCheckBoxMenuItem("Auto Generate", autoGenerate );
	autoGenerateMenu.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// GENERATE
			JCheckBoxMenuItem m = (JCheckBoxMenuItem)e.getSource();
			setAutoGeneratorState(m.getState());
			}
		});
	
	generateMenu.add(autoGenerateMenu);
	
	autoSaveMenu = new JCheckBoxMenuItem("Auto Save On Exit", autoSave );
	autoSaveMenu.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// SAVE ALL PUZZLES ON EXIT
			JCheckBoxMenuItem m = (JCheckBoxMenuItem)e.getSource();
			autoSave = m.getState();
			try
	            {
	            generator.save();
	            }
            catch ( FileNotFoundException exc )
	            {
	            exc.printStackTrace();
	            JOptionPane.showMessageDialog(SudokuGUI.this, "File store failed to save due to:\n"+exc.getMessage(), "Save failed", JOptionPane.ERROR_MESSAGE);
	            }
			}
		});
	
	generateMenu.add(autoSaveMenu);

	generateMenu.add(new JSeparator());

	JMenuItem mItem = new JMenuItem("Empty Puzzle");
	generateMenu.add(mItem);
	
	mItem.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// EMPTY PUZZLE
			Puzzle p = sudokuSolver.getPuzzle();
			p.clear();
			
		    try
			    {
			    if ( currentFile != null && currentFile.isFile() )
			    	currentFile = currentFile.getAbsoluteFile().getParentFile();
			    solutionsPanel.clear();
			    gridPanel.setEmptyPuzzle(p);
			    frameType.setTitle("Sudoku");
			    }
		    catch ( IllegalGridException exc )
			    {
			    // this should never happen as the grid should be generated and hence ok
			    exc.printStackTrace();
			    }

			}
		});
	
	addGenerateMenuItem(generateMenu, PuzzleDifficulty.BEGINNER);
	addGenerateMenuItem(generateMenu, PuzzleDifficulty.EASY);
	addGenerateMenuItem(generateMenu, PuzzleDifficulty.MEDIUM);
	addGenerateMenuItem(generateMenu, PuzzleDifficulty.HARD);
	addGenerateMenuItem(generateMenu, PuzzleDifficulty.EXTREME);

	return generateMenu;
	}

/**
 * Adds a menu item
 * 
 * @param gen the menu to add the menu item too
 * @param dif the puzzle difficulty of the menu item 
 */
private void addGenerateMenuItem(JMenu gen, final PuzzleDifficulty dif)
	{
	JMenuItem mItem = new JMenuItem(dif.toString());
	mItem.setForeground(Color.RED);
	menuMap.put(dif, mItem);
	gen.add(mItem);
	
	mItem.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// DISPALY PUZZLE
		    if ( currentFile != null && currentFile.isFile() )
		    	currentFile = currentFile.getAbsoluteFile().getParentFile();
			generateMenu.setEnabled(false);

			// Set a progress monitor to display after 300 milliseconds if a puzzle
			// has not been generated 
			ProgressMonitor progressMonitor = new ProgressMonitor(SudokuGUI.this,
					"Generating a Sudoku puzzle. Difficulty = "+dif.toString()+".             ",
					" ", 0, MAX_PROGRESS+1);
			progressMonitor.setMillisToDecideToPopup(300);
			progressMonitor.setProgress(0);
			
			GeneratePuzzleTask task = new GeneratePuzzleTask(dif, progressMonitor);
			task.execute();		
			}
		});
	}

/**
 * Creates the Options menu
 * 
 * @return the menu 
 */
private JMenu createOptionsMenu()
	{
	JMenu optsMenu = new JMenu("Options");
	
	autoSolveMenu = new JCheckBoxMenuItem("Auto Solve Puzzle", autoSolve);
	autoSolveMenu.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// AUTO SOLVE
			JCheckBoxMenuItem m = (JCheckBoxMenuItem)e.getSource();
			autoSolve = m.getState();
			
			solutionsPanel.setAutoSolve(autoSolve);
			
			if ( autoSolve )
				{
	            try
	                {
	                solutionsPanel.showSolution(new Puzzle(gridPanel.getDisplayedValues(), sudokuSolver.getPuzzle().getDifficulty()));
	                }
                catch ( IllegalGridException exc )
	                {
	                exc.printStackTrace();
	                JOptionPane.showMessageDialog(SudokuGUI.this, "This puzzle contains errors:\n"+exc.getMessage(), "Solution Error", JOptionPane.ERROR_MESSAGE);
	                }
				}
            else
			    solutionsPanel.clear();
			}
		});
	
	optsMenu.add(autoSolveMenu);
	
	showNumbersMenu = new JCheckBoxMenuItem("Show Possible Numbers", showPossibleNums);
	showNumbersMenu.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// SHOW POSSIBLE NUMBERS
			JCheckBoxMenuItem m = (JCheckBoxMenuItem)e.getSource();
			showPossibleNums = m.getState();
			
			gridPanel.showPossibleNumbers(showPossibleNums);
			}
		});
	
	optsMenu.add(showNumbersMenu);
	optsMenu.add(new JSeparator());
	
	restoreOnStartUpMenu = new JCheckBoxMenuItem("Restore Existing Puzzle on Start Up", restoreOnStartUp);
	restoreOnStartUpMenu.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// Restore application state on startup
			JCheckBoxMenuItem m = (JCheckBoxMenuItem)e.getSource();
			restoreOnStartUp = m.getState();
			}
		});
	
	optsMenu.add(restoreOnStartUpMenu);

	return optsMenu;
	}

/**
 * Creates the Snapshot menu
 * 
 * @return the menu 
 */
private JMenu createSnapshotMenu()
	{
	final JSeparator menuSeparator = new JSeparator();
	final JMenu snapshotMenu = new JMenu("Snapshot");
	
	final ActionListener snapshotListener = new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// SHOW SNAPSHOT
			Snapshot snap = (Snapshot)e.getSource();
			
		    solutionsPanel.clear();
		    try
	            {
	            gridPanel.setPuzzle(snap.snap);
	            }
            catch ( IllegalGridException exc )
	            {
	            // This shouldn't happen as we have saved a legal puzzle
	            exc.printStackTrace();
	            }
			}
		};
	JMenuItem mItem = new JMenuItem("Add Snapshot");
	snapshotMenu.add(mItem);
	
	mItem.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// ADD SNAPSHOT
			
			// if there's too many then remove the earliest ones
			while ( snapshots.size() >= MAX_SNAPSHOTS )
				{
				Snapshot firstSnap = snapshots.remove(0);
				snapshotMenu.remove(firstSnap);
				}
			
			Puzzle p = new Puzzle(gridPanel.getDisplayedValues(), sudokuSolver.getPuzzle().getDifficulty());
			
			DateFormat df = DateFormat.getTimeInstance(DateFormat.MEDIUM);
			String time = df.format(new Date());
			Snapshot snap = new Snapshot("Snapshot at "+time, p);
			snap.addActionListener(snapshotListener);
			
			if ( snapshots.size() == 0 )
				snapshotMenu.add(menuSeparator);

			snapshots.add(snap);
			snapshotMenu.add(snap);
			}
		});
	
	mItem = new JMenuItem("Remove All Snapshots");
	snapshotMenu.add(mItem);
	
	mItem.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// REMOVE ALL SNAPSHOTS
			for ( Snapshot snap : snapshots )
	            {
				snapshotMenu.remove(snap);
	            }

			snapshotMenu.remove(menuSeparator);
			snapshots.clear();
			}
		});
	
	
	
	return snapshotMenu;
	}

/**
 * Creates the About menu item
 * 
 * @return the menu item
 */
private JMenuItem createAboutMenu()
	{
	JMenuItem aboutMenu = new JMenuItem("About");
	
	aboutMenu.addActionListener(new ActionListener()
		{
		@Override
		public void actionPerformed(ActionEvent e)
			{
			// ABOUT
            JOptionPane.showMessageDialog(SudokuGUI.this, "Sudoku Solver\nVersion: 0.5 Beta\nReleased: 23/08/2010", "About", JOptionPane.INFORMATION_MESSAGE);
			}
		});
	
	return aboutMenu;
	}

/**
 * Inits the panel
 */
private void initPanel()
	{
	setLayout(new BorderLayout());
	JPanel p = new JPanel();
	add(p, BorderLayout.CENTER);
	
	p.setBorder(new EmptyBorder(5, 5, 5, 5));
	p.setLayout(new GridLayout(1,2,5,5));
	
	// add the sudoku panel
	gridPanel = new SudokuGrid(sudokuSolver);
	gridPanel.showPossibleNumbers(showPossibleNums);
	gridPanel.setPreferredSize(new Dimension(399, 405));
	p.add(gridPanel);
	
	// add the events solution panel
	solutionsPanel = new SolutionPanel(sudokuSolver, sudokuSolver.getSequenceBuffer(), gridPanel);
	solutionsPanel.setAutoSolve(autoSolve);
	p.add(solutionsPanel);
	}


/**
 * Notification that this window has been closed (or in the case of an applet the page is no longer
 * being displayed)
 */
public void windowClosed()
	{
	generator.stopBackgroundGeneration();
	
	ApplicationMemento mem = new ApplicationMemento();
	try
	    {
	    mem.save();
	    }
    catch ( FileNotFoundException exc1 )
	    {
	    // Nothing we can do at this point
	    exc1.printStackTrace();
	    }
	
	if ( autoSave )
		{
		try
	        {
//	        System.out.println("Saving puzzles on exit");
	        generator.save();
	        }
        catch ( FileNotFoundException exc )
	        {
	        // Nothing we can do at this point
	        exc.printStackTrace();
	        }
		}
	}

@Override
public void itemStored(PriorityQueue<PuzzleStoreItem> store, PuzzleStoreItem item)
	{
	if ( store.size() > 0 )
		{
		PuzzleDifficulty d = item.getPuzzle().getDifficulty();
		JMenuItem menu = menuMap.get(d);
		if ( menu != null )
			{
			menu.setForeground(Color.BLACK);
			menu.setText(d+" ("+store.size()+")");
			}
		}
	}


@Override
public void itemRetrieved(PriorityQueue<PuzzleStoreItem> store, PuzzleStoreItem item)
	{
	PuzzleDifficulty d = item.getPuzzle().getDifficulty();
	JMenuItem menu = menuMap.get(d);
	
	if ( menu == null )
		return;
	
	if ( store.size() == 0 )
		{
		menu.setForeground(Color.RED);
		}
	else
		{
		menu.setText(d+" ("+store.size()+")");
		}
	}

/**
 * Sets the auto generate state
 * 
 * @param state the state 
 */
private void setAutoGeneratorState(boolean state)
	{
	autoGenerate = state;
	
	if ( autoGenerate )
		generator.startBackgroundGeneration();
	else
		generator.stopBackgroundGeneration();
	}

/**
 * @param args
 */
public static void main(String[] args)
	{
    try
        {
        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        }
    catch ( ClassNotFoundException e )
        {
        System.out.println("Failed to set Look and Feel "+e);
        }
    catch ( InstantiationException e )
        {
        System.out.println("Failed to set Look and Feel "+e);
        }
    catch ( IllegalAccessException e )
        {
        System.out.println("Failed to set Look and Feel "+e);
        }
    catch ( UnsupportedLookAndFeelException e )
        {
        System.out.println("Failed to set Look and Feel "+e);
        }

	
	if ( args.length != 81 || 
		(args.length == 1 && args[0].length() != 81) )
		{
//		System.out.println("There must be 81 numbers/spaces, you have entered "+args.length);
		
		// all types req'd
//		args = new String[]{"    358   1 7  9  6.7       2 8   5  3  6  8  4   2 6       4 8  2  9 1   138    "};
		
//		args = new String[]{"   2 1  4  8 9 3   4 6   9 51  7 2   6   3 7 4 2    31 3   2 4  8475 1     3 6 2 "};
//		args = new String[]{"1 64      2  3 6         89 7   4  34   2   76  9   5 29         1 9  7      51 4"};
		
		// split group and associated pair solver req'd
//		args = new String[]{"6    4357  8 6           2 1   7 83 8       1 54 8   2 6           4 7  7951    4"};

//		args = new String[]{" 4  5   3        8   6   5   61  53 71  6 2 42  3 86   5    3 6  7     56 8  7941"}; 
//		return;
		}

	SudokuGUI.show(args);
	}


//=========================================================================================================
class GeneratePuzzleTask extends SwingWorker<Puzzle, PropertyChangeEvent> implements PropertyChangeListener
	{
	private final PuzzleDifficulty dif;
	private final ProgressMonitor progMon;
	private int progress = 0;
	
	public GeneratePuzzleTask(PuzzleDifficulty d, ProgressMonitor pm)
		{
		dif = d;
		progMon = pm;
		}
	
	@Override
	public Puzzle doInBackground()
		{
		Puzzle p = generator.generatePuzzle(dif, this);
		return p;
		}
	
	@Override
	public void done()
		{
		// close the progress monitor window
		progMon.close();
		
		// re-enable the menu item
		generateMenu.setEnabled(true);
		
		if ( isCancelled() )
			return;
		
		Puzzle p;
	    try
		    {
		    // get the generated puzzle
		    p = get();
		    }
	    catch ( InterruptedException exc1 )
		    {
		    exc1.printStackTrace();
		    return;
		    }
	    catch ( ExecutionException exc1 )
		    {
		    exc1.printStackTrace();
		    return;
		    }
	    
	    try
		    {
		    gridPanel.setPuzzle(p);
		    }
	    catch ( IllegalGridException exc )
		    {
		    // this should never happen as the grid should be generated and hence ok
		    exc.printStackTrace();
		    }
		}
	
	@Override
	protected void process(List<PropertyChangeEvent> chunks)
		{
		PropertyChangeEvent evt = chunks.get(chunks.size() - 1);
		int i = (Integer)evt.getNewValue();
		progress++;
		
		if ( progress > MAX_PROGRESS )
			progress = 1;
		
		progMon.setProgress(progress);
	
	    String message = "Iteration "+(i+1)+" failed to produce puzzle.\nTrying again ...";
	    progMon.setNote(message);
	    
		if ( progMon.isCanceled() ) 
			{
			generator.cancelGenerate();
			cancel(false);
			}
		}
	
	@Override
	public void propertyChange(PropertyChangeEvent evt)
		{
		if ("progress" == evt.getPropertyName() && !isCancelled() ) 
			{
			publish(evt);
			}
		}
	}

/**
 * A snapshot of the current puzzle
 * 
 * @author A.G.Docherty
 */
private class Snapshot extends JMenuItem
	{
	private static final long serialVersionUID = -7046305340497168315L;
	private final Puzzle snap;
	
	Snapshot(String name, Puzzle p)
		{
		super(name);
		snap = p;
		}
	}

private class ApplicationMemento
{
private static final String APP_SAVE_FNAME = "sudokuState.psdk";
private static final String AUTO_SOLVE = "AutoSolve";
private static final String AUTO_SAVE = "AutoSave";
private static final String AUTO_GENERATE = "AutoGenerate";
private static final String SHOW_ALL_NUMBERS = "ShowAllNumbers";
private static final String FRAME_SIZE_WIDTH = "FrameWidth";
private static final String FRAME_SIZE_HEIGHT = "FrameHeight";
private static final String FRAME_LOCATION_X = "FrameX";
private static final String FRAME_LOCATION_Y = "FrameY";
private static final String NUM_DLG_SIZE_WIDTH = "NumDlgWidth";
private static final String NUM_DLG_SIZE_HEIGHT = "NumDlgHeight";
private static final String SOLVER_INDEX = "SolverIndex";
private static final String RESTORE_ON_STARTUP = "RestoreOnStartUp";

private boolean solve;
private boolean save;
private boolean generate;
private boolean showNumbers;
private Puzzle puzzle;
private int solverIndex;
private int frameWidth;
private int frameHeight;
private int frameX;
private int frameY;
private int numDlgWidth;
private int numDlgHeight;
private boolean restore;

public ApplicationMemento()
	{
	// get the option values
	solve = autoSolve;
	save = autoSave;
	generate = autoGenerate;
	showNumbers = showPossibleNums;
	restore = restoreOnStartUp;
	
	// get the screen size and location
	Dimension frameSize = frameType.getSize();
	frameWidth = frameSize.width;
	frameHeight = frameSize.height;
	
	Point frameLocation = frameType.getLocation();
	frameX = frameLocation.x;
	frameY = frameLocation.y;

	// get the number dialog size
	Dimension numDlgSize = NumberInputDialog.getPanelSize();
	numDlgWidth = numDlgSize.width;
	numDlgHeight = numDlgSize.height;

	// Capture the current screen state 
	// ie if solver is running get the initial grid from the SequenceBuffer and the currently
	// selected index from the solver list
	// else capture the currently displayed grid
	if ( autoSolve )
		{
		puzzle = seqBuff.getStartGrid();
		solverIndex = solutionsPanel.getSelectedIndex();
		}
	else
		{
		puzzle = new Puzzle(gridPanel.getDisplayedValues(), PuzzleDifficulty.UNKNOWN);
		solverIndex = 0;
		}
	}

public void restoreValues()
	{
	// set the option values
	autoSolve = solve;
	autoSolveMenu.setSelected(autoSolve);
	solutionsPanel.setAutoSolve(autoSolve);

	autoSave = save;
	autoSaveMenu.setSelected(autoSave);
	
	autoGenerateMenu.setSelected(generate);
	setAutoGeneratorState(generate);
	
	showPossibleNums = showNumbers;
	showNumbersMenu.setSelected(showPossibleNums);
	gridPanel.showPossibleNumbers(showPossibleNums);
	
	restoreOnStartUp = restore;
	restoreOnStartUpMenu.setSelected(restoreOnStartUp);
	
	// only restore the screen setting if restore on startup is checked
	if ( !restoreOnStartUp )
		return;
	
	// set the screen size and location
	frameType.setSize(new Dimension(frameWidth, frameHeight));
	frameType.setLocation(new Point(frameX, frameY));

	// set the number dialog size
	NumberInputDialog.setPanelSize(new Dimension(numDlgWidth, numDlgHeight));

	// set the current screen state 
	// ie if solver is running set the initial grid and the currently
	// selected index from the solver list
	// else set the currently displayed grid
    try
	    {
	    gridPanel.setPuzzle(puzzle);
	    }
    catch ( IllegalGridException exc )
	    {
	    // this should never happen as the grid should be generated and hence ok
	    exc.printStackTrace();
	    }

	if ( autoSolve )
		{
		// do this later as the screen needs to redraw to ensure the selected index will be visible
	    EventQueue.invokeLater(new Runnable()
			{
			@Override
			public void run()
				{
				solutionsPanel.setSelectedIndex(solverIndex);
				}
		});
		}
	}

/**
 * Saves the current state of the application 
 * @throws FileNotFoundException 
 */
void save() throws FileNotFoundException
	{
	// open the file
	File file = new File(APP_SAVE_FNAME);
	PrintWriter w = new PrintWriter(file);
	
	try {
		write(w);
		}
	finally
		{
		w.close();
		}
	}

/**
 * Writes the application data to the output stream
 * 
 * @param w the writer to write to
 */
private void write(PrintWriter w)
	{
    w.println(RESTORE_ON_STARTUP+":"+restore);
    w.println(AUTO_SOLVE+":"+solve);
    w.println(AUTO_SAVE+":"+save);
    w.println(AUTO_GENERATE+":"+generate);
    w.println(SHOW_ALL_NUMBERS+":"+showNumbers);
    w.println(FRAME_SIZE_WIDTH+":"+frameWidth);
    w.println(FRAME_SIZE_HEIGHT+":"+frameHeight);
    w.println(FRAME_LOCATION_X+":"+frameX);
    w.println(FRAME_LOCATION_Y+":"+frameY);
    w.println(NUM_DLG_SIZE_WIDTH+":"+numDlgWidth);
    w.println(NUM_DLG_SIZE_HEIGHT+":"+numDlgHeight);
    w.println(SOLVER_INDEX+":"+solverIndex);
    
    w.println("=====================");
    puzzle.write(w);
	}

/**
 * Loads the previously saved state of the application 
 * 
 * @param r the reader to read from
 * @throws IOException 
 * @throws IllegalGridException 
 */
void load() throws IOException, IllegalGridException
	{
	BufferedReader r = new BufferedReader(new FileReader(APP_SAVE_FNAME));

	try 
		{
		read(r);
		}
	finally
		{
		r.close();
		}
	}

/**
 * Reads this file from the input stream
 * 
 * @param r the reader to read from
 * @throws IOException 
 * @throws IllegalGridException 
 */
private void read(BufferedReader r) throws IOException, IllegalGridException
	{
	String dif = r.readLine();
	
	while ( dif != null && !dif.startsWith("===") )
		{
		// LOAD THE OPTIONS
		String[] parts = dif.split(":");
		
		if ( parts.length != 2 )
			{
			System.err.println("Loading application data: ERROR '"+dif+"' is not a valid line");
			continue;
			}
		
		if ( parts[0].equals(RESTORE_ON_STARTUP) )
			{
			restore = parseBoolean(parts[1]);
			}
		else if ( parts[0].equals(AUTO_SOLVE) )
			{
			solve = parseBoolean(parts[1]);
			}
		else if ( parts[0].equals(AUTO_SAVE) )
			{
			save = parseBoolean(parts[1]);
			}
		else if ( parts[0].equals(AUTO_GENERATE) )
			{
			generate = parseBoolean(parts[1]);
			}
		else if ( parts[0].equals(SHOW_ALL_NUMBERS) )
			{
			showNumbers = parseBoolean(parts[1]);
			}
		else if ( parts[0].equals(FRAME_SIZE_WIDTH) )
			{
			frameWidth = parseInt(parts[1]);
			}
		else if ( parts[0].equals(FRAME_SIZE_HEIGHT) )
			{
			frameHeight = parseInt(parts[1]);
			}
		else if ( parts[0].equals(FRAME_LOCATION_X) )
			{
			frameX = parseInt(parts[1]);
			}
		else if ( parts[0].equals(FRAME_LOCATION_Y) )
			{
			frameY = parseInt(parts[1]);
			}
		else if ( parts[0].equals(NUM_DLG_SIZE_WIDTH) )
			{
			numDlgWidth = parseInt(parts[1]);
			}
		else if ( parts[0].equals(NUM_DLG_SIZE_HEIGHT) )
			{
			numDlgHeight = parseInt(parts[1]);
			}
		else if ( parts[0].equals(SOLVER_INDEX) )
			{
			solverIndex = parseInt(parts[1]);
			}

		dif = r.readLine();
		}

	String lineSeparator = System.getProperty("line.separator", "\n");
	StringBuilder sb = new StringBuilder();

	dif = r.readLine();

	while ( dif != null )
		{
		// LOAD THE PUZZLE
		sb.append(dif);
		sb.append(lineSeparator);
		
		dif = r.readLine();
		}
	
	BufferedReader sr = new BufferedReader(new StringReader(sb.toString()));
	puzzle = new Puzzle(sr);
	}

private int parseInt(String val)
	{
	int i = 0;
	try
	    {
	    i = Integer.parseInt(val);
	    }
    catch ( NumberFormatException exc )
	    {
	    System.err.println("The value '"+val+"' isn't a valid integer");
	    }
    
	return i;
	}

private boolean parseBoolean(String val)
	{
	boolean b = Boolean.parseBoolean(val);
	return b;
	}
}

}
