/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.border.LineBorder;

import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.Puzzle.PuzzleDifficulty;
import uk.co.keang.sudoku.SudokuButton.BtnState;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.swingaddons.VariableGridLayout;

/**
 * The sudoku panel consisting of n sub-panels of n x n SudokuButtons 
 * 
 * @author A.G.Docherty
 *
 */
class SudokuGrid extends JPanel 
{
private static final long serialVersionUID = 9120427056774861843L;
private static final int BORDER_SIZE = 1;
private final SudokuSolver sudoku;
private final Block[] blocks = new Block[SudokuSolver.GRID_SIZE];
private SudokuButton[][] btns;
private final List<Point> currentSelected = new ArrayList<Point>();
private final List<Point> currentReferenced = new ArrayList<Point>();
private final List<GridChangeListener> listeners = new ArrayList<GridChangeListener>();
private boolean initialPuzzle;
private boolean emptyPuzzle;


public SudokuGrid(SudokuSolver puzzle)
	{
	this.sudoku = puzzle;
	
	createButtons();
	initPanel();
	setDisplayedValues(sudoku.getPuzzle());
	}

private void createButtons()
	{
	btns = new SudokuButton[SudokuSolver.GRID_SIZE][SudokuSolver.GRID_SIZE];

	ActionListener btnListener = new ActionListener()
				{
				@Override
				public void actionPerformed(ActionEvent e)
					{
					SudokuButton btn = (SudokuButton)e.getSource();
					
					Cell cell = new Cell(btn.getPossibilites());
					
					if ( emptyPuzzle && cell.size() == SudokuSolver.GRID_SIZE )
						cell.clear();

					Cell nums = NumberInputDialog.show(btn, cell);
					
					// if ok pressed and something has changed
					if ( nums != null && !nums.equals(cell) )
						{
                        if ( nums.size() == 0 )
                        	{
                        	if ( !emptyPuzzle )
                        		{
	                        	String msg = "You must select at least 1 number (or select all numbers to clear the cell)";
	                        	JOptionPane.showMessageDialog(btn, msg, "Illegal Number", JOptionPane.ERROR_MESSAGE);
                        		}
                        	else
                        		{
                        		nums.addAll(Puzzle.allPossibleNumbers);
    							btn.setPossibilites(nums);
		                        try
	                                {
	                                fireGridChange(getLocation(btn), cell, nums);
	                                }
                                catch ( IllegalGridException exc )
	                                {
	                                // we are clearing the cell so it can't be illegal
	                                exc.printStackTrace();
	                                }
                        		}
                        	}
                        else
                        	{
							nums.userEntered = nums.size() < SudokuSolver.GRID_SIZE;
							
							btn.setPossibilites(nums);
							try
		                        {
		                        fireGridChange(getLocation(btn), cell, nums);
		                        
		                        Puzzle p = new Puzzle(getDisplayedValues(), PuzzleDifficulty.UNKNOWN);
		                        if ( !p.isLegal() )
		                        	throw new IllegalGridException(currentSelected);
		                        }
	                        catch ( IllegalGridException exc )
		                        {
		                        String msg;
		                        if ( nums.size() == 1 )
		                        	msg = "The number "+Puzzle.getNumbersAsText(nums)+" is not valid for this cell";
		                        else
		                        	msg = "The numbers "+Puzzle.getNumbersAsText(nums)+" do not include a valid number for this cell";	
							    JOptionPane.showMessageDialog(btn, msg, "Illegal Number", JOptionPane.ERROR_MESSAGE);
								
								btn.setPossibilites(cell);
		                        try
	                                {
	                                fireGridChange(getLocation(btn), nums, cell);
	                                }
                                catch ( IllegalGridException exc1 )
	                                {
	                                // this shouldn't happen as we are setting the value back to what it was
	                                exc1.printStackTrace();
	                                }
		                        }
                        	}
						}			
					}

				private Point getLocation(SudokuButton btn)
	                {
	                for ( int y = 0; y < btns.length; y++ )
	                    {
	                    for ( int x = 0; x < btns[y].length; x++ )
	                        {
	                        if ( btns[y][x] == btn )
	                        	return new Point(x, y);
	                        }
	                    }

	                return null;
	                }
				};
				
	for ( int y = 0; y < btns.length; y++ )
	    {
	    for ( int x = 0; x < btns[y].length; x++ )
	        {
	        btns[y][x] = new SudokuButton(x, y);
	        btns[y][x].addActionListener(btnListener);
	        }
	    }
	}

private void initPanel()
	{
	setLayout(new VariableGridLayout(2, 2, 5, 0));
	
	JLabel blank = new JLabel(" ");
	blank.setPreferredSize(new Dimension(1, 1));
	add(blank);
	
	JPanel top = new JPanel();
	top.setLayout(new GridLayout(1, SudokuSolver.SUB_GRID_SIZE, BORDER_SIZE*2, 0));
	
	String[] lblTextXAxis = Puzzle.getXAxisLabels();
	for ( int i = 0; i < SudokuSolver.SUB_GRID_SIZE; i++ )
	    {
	    JPanel p = new JPanel();
		p.setLayout(new GridLayout(1, SudokuSolver.SUB_GRID_SIZE, 0, 0));
		for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
		    {
		    int num = j+(i*SudokuSolver.SUB_GRID_SIZE);
		    JLabel lbl = new JLabel(lblTextXAxis[num], JLabel.CENTER);
		    Dimension d = lbl.getPreferredSize();
		    d.height /= 2;
		    lbl.setPreferredSize(d);
		    p.add(lbl);
		    }
		
	    top.add(p);
	    }
	
	add(top);

	JPanel side = new JPanel();
	side.setLayout(new GridLayout(SudokuSolver.SUB_GRID_SIZE, 1, BORDER_SIZE*2, 0));
	
	String[] lblTextYAxis = Puzzle.getYAxisLabels();
	for ( int i = 0; i < SudokuSolver.SUB_GRID_SIZE; i++ )
	    {
	    JPanel p = new JPanel();
		p.setLayout(new GridLayout(SudokuSolver.SUB_GRID_SIZE, 1, 0, 0));
		for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
		    {
		    int num = j+(i*SudokuSolver.SUB_GRID_SIZE);
		    JLabel lbl = new JLabel(" "+lblTextYAxis[num], JLabel.CENTER);
		    p.add(lbl);
		    }
		
	    side.add(p);
	    }
	
	add(side);

	
	JPanel puzzle = new JPanel();
	puzzle.setBorder(new LineBorder(Color.BLACK, BORDER_SIZE));
	puzzle.setLayout(new GridLayout(SudokuSolver.SUB_GRID_SIZE, SudokuSolver.SUB_GRID_SIZE, 0, 0));
	
	for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
	    {
	    blocks [i] = new Block(i);
	    puzzle.add(blocks[i]);
	    }
	
	add(puzzle);
	}

Cell[][] getDisplayedValues()
	{
    Cell[][] grid = new Cell[SudokuSolver.GRID_SIZE][SudokuSolver.GRID_SIZE];
	
	for ( int y = 0; y < grid.length; y++ )
	    {
	    for ( int x = 0; x < grid[y].length; x++ )
	        {
	        grid[y][x] = new Cell(btns[y][x].getPossibilites());
	        }
	    }
	
	return grid;
	}

/**
 * Sets a new puzzle
 * 
 * @param puz
 * @throws IllegalGridException 
 */
void setPuzzle(Puzzle puz) throws IllegalGridException
	{
	emptyPuzzle = false;
	clear();
	setDisplayedValues(puz);
	fireGridReplaced(puz);
	}

/**
 * Sets an empty puzzle
 * 
 * @param puz
 * @throws IllegalGridException 
 */
void setEmptyPuzzle(Puzzle puz) throws IllegalGridException
	{
	setPuzzle(puz);
	emptyPuzzle = true;
	}

/**
 * Sets the changed puzzle values to display
 * 
 * @param puz
 */
void setDisplayedValues(Puzzle puz)
	{
	Cell[][] grid = puz.getGrid();
	for ( int y = 0; y < grid.length; y++ )
	    {
	    for ( int x = 0; x < grid[y].length; x++ )
	        {
	        btns[y][x].setPossibilites(grid[y][x]);
	        btns[y][x].repaint();
	        }
	    }
	
	revalidate();
	repaint();
	}

public void setPossibleValues(int x, int y, Cell values)
	{
	btns[y][x].setPossibilites(values);
    btns[y][x].repaint();
	}

public void clear()
	{
	for ( Point p : currentSelected )
	    {
	    btns[p.y][p.x].setDisplayState(BtnState.NORMAL);
	    btns[p.y][p.x].repaint();
	    }
	
	for ( Point p : currentReferenced )
	    {
	    btns[p.y][p.x].setDisplayState(BtnState.NORMAL);
	    btns[p.y][p.x].repaint();
	    }
	}

public void setResolvedValues(List<Point> locs, Cell[][] sets)
	{
	for ( Point p : locs )
	    {
		btns[p.y][p.x].setResolved(sets[p.y][p.x].iterator().next());
	    btns[p.y][p.x].repaint();
	    }
	}

public void setSelectedCells(List<Point> locs, Cell removed)
	{
	currentSelected.clear();
	
	if ( locs == null )
		return;
	
	currentSelected.addAll(locs);
	
	for ( Point p : currentSelected )
	    {
	    btns[p.y][p.x].setDisplayState(BtnState.SELECTED);
	    btns[p.y][p.x].setRemovePossibilites(removed);
		
	    btns[p.y][p.x].repaint();
	    }
	}

public void setReferenced(List<Point> locs, Cell removed)
	{
	currentReferenced.clear();
	
	if ( locs == null )
		return;
	
	currentReferenced.addAll(locs);
	
	for ( Point p : currentReferenced )
	    {
	    btns[p.y][p.x].setDisplayState(BtnState.REFERENCE);
		btns[p.y][p.x].setRemovePossibilites(removed);
	    btns[p.y][p.x].repaint();
	    }
	}

/**
 * Sets the show possible numbers state
 * 
 * @param show true to show all possible numbers, false to just show those entered by the user
 */
void showPossibleNumbers(boolean show)
	{
	for ( int y = 0; y < btns.length; y++ )
	    {
	    for ( int x = 0; x < btns[y].length; x++ )
	        {
	        btns[y][x].showPossibleNumbers(show);
	        btns[y][x].repaint();
	        }
	    }
	}

/**
 * Sets whether the grid is displaying the initial puzzle or not. This is determined
 * solely by the index of the selected solution event ie if it is 0 then this is true
 * else it is false.
 * 
 * @param initial  true if this is showing the initial puzzle 
 */
void setInitialGrid(boolean initial)
	{
	if ( initialPuzzle == initial )
		return;
	
	initialPuzzle = initial;
	}


/**
 * Adds a listener. 
 *
 * @param l - the listener to add
 */
public void addGridChangeListener(GridChangeListener l)
	{
	listeners.add(l);
	}

/**
 * Removes the listener
 *
 * @param l - the listener to remove
 */
public void removeGridChangeListener(GridChangeListener l)
	{
	listeners.remove(l);
	}

/**
 * Removes the listener
 *
 * @param l - the listener to remove
 */
public void removeAllGridChangeListeners()
	{
	listeners.clear();
	}

/**
 * fires the Grid change event to all listeners
 * @param nums 
 * @param cell 
 * @param point 
 * @throws IllegalGridException 
 */
protected void fireGridChange(Point loc, Cell origVal, Cell newVal) throws IllegalGridException
	{
	GridChangeEvent e = new GridChangeEvent(loc, origVal, newVal);
	
	// call all listeners
	for ( GridChangeListener listener : listeners )
		{
		listener.gridChanged(e);
		}
	}

/**
 * fires the Grid change event to all listeners
 * @param nums 
 * @param cell 
 * @param point 
 * @throws IllegalGridException 
 */
protected void fireGridReplaced(Puzzle p) throws IllegalGridException
	{
//	GridChangeEvent e = new GridChangeEvent(loc, origVal, newVal);
	
	// call all listeners
	for ( GridChangeListener listener : listeners )
		{
		listener.gridReplaced(p);
		}
	}



//========================================================================================================
private class Block extends JPanel
{
private static final long serialVersionUID = 5924265091421835368L;

Block(int i)
	{
	setLayout(new GridLayout(SudokuSolver.SUB_GRID_SIZE, SudokuSolver.SUB_GRID_SIZE, 1, 1));
	setBorder(new LineBorder(Color.BLACK, BORDER_SIZE));

	int y = (i / SudokuSolver.SUB_GRID_SIZE) * SudokuSolver.SUB_GRID_SIZE;
	int x = (i % SudokuSolver.SUB_GRID_SIZE) * SudokuSolver.SUB_GRID_SIZE;
	
	for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
	    {
		for ( int k = 0; k < SudokuSolver.SUB_GRID_SIZE; k++ )
		    {
		    add(btns[y+j][x+k]);
		    }
	    }
	
	}
}


}
