/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.util.ArrayList;
import java.util.List;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SolverListener;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.SolutionEvent;

/**
 * An abstract sudouku solver pattern
 *  
 * @author A.G.Docherty
 */
public abstract class AbstractSolver
{
public enum SolverDifficulty
	{
	UNKNOWN,
	BEGINNER,
	EASY,
	MEDIUM,
	HARD,
	EXTREME;
	};

protected int found = 0;
private final List<SolverListener> listeners = new ArrayList<SolverListener>();
protected final Puzzle puzzle;
protected final Cell[][] hGrid;
private int diffFactor = 1;
private final SolverDifficulty difficulty;

/**
 * Constructs a solver
 *  
 * @param hGrid
 */
AbstractSolver(Puzzle p, int factor, SolverDifficulty dif)
	{
	puzzle = p;
	
	if ( factor > 0 )
		diffFactor = factor;
	
	difficulty = dif;
	
	hGrid = puzzle.getGrid();
	}

/**
 * Runs the solver once to try to resolve individual squares
 *  
 * @return the number of solutions found
 * @throws IllegalGridException 
 */
public abstract int findSolutions() throws IllegalGridException ;

/**
 * Gets the minimum difficulty of this solver
 * 
 * @return the difficulty
 */
public SolverDifficulty getSolverDifficulty()
	{
	return difficulty;
	}

/**
 * Adds a solver listener. 
 *
 * @param l - the listener to add
 */
public void addSolverListener(SolverListener l)
	{
	listeners.add(l);
	}

/**
 * Removes the solver listener
 *
 * @param l - the listener to remove
 */
public void removeSolverListener(SolverListener l)
	{
	listeners.remove(l);
	}

/**
 * Removes all the solver listeners
 *
 * @param l - the listener to remove
 */
public void removeAllSolverListeners()
	{
	listeners.clear();
	}

/**
 * True if there are registered solver listeners
 * 
 * @return true of there are solver listeners
 */
protected boolean hasSolverListeners() { return listeners.size() > 0; }

/**
 * fires the possibilities reduced event to all listeners
 */
protected void fireSolutionEvent(SolutionEvent event)
	{
	if ( listeners.size() == 0 )
		return;
	
	// call all listeners
	for ( SolverListener listener : listeners )
		{
		listener.removedPossibleValues(event);
		}
	}
/**
 * Removes the values from the possible values 
 *  
 * @param possibleVals the set of possible values
 * @param removeVals the values to remove
 * @return the values removed from this set
 */
protected final Cell removeAllFromSet(Cell possibleVals, Cell removeVals)
	{
	Cell common = new Cell();

	if ( possibleVals.size() == 1 )
		return common;
	
	// find which values are in both sets
	// this way is marginally less efficient than below
//	Cell copy = new HashCell(possibleVals);
//	copy.retainAll(removeVals);
	

	// find which values are common to both sets
    for ( Integer val : removeVals )
	    {
	    if ( possibleVals.contains(val) )
	        common.add(val);
	    }

	// remove the common values
	if ( common.size() > 0 )
		{
		// remove the values
		possibleVals.removeAll(common);
		}

	return common;
	}

/**
 * Retains these values removing all other possible values 
 *  
 * @param possibleVals the set of possible values
 * @param removeVals the values to remove
 * @return the values removed from this set
 */
protected final Cell retainAllInSet(Cell possibleVals, Cell retainVals)
	{
	// find which values are in both sets
	Cell removed = new Cell(possibleVals);
	removed.removeAll(retainVals);

	// fire a change event
	if ( removed.size() > 0 )
		{
		// remove the values
		possibleVals.retainAll(retainVals);
		}

	return removed;
	}

/**
 * Gets a difficulty factor for this solver where the higher the number the higher the level of 
 * expertise required to solve the puzzle
 * 
 * @return the difficulty factor
 */
public int getDifficultyFactor()
	{
	return diffFactor;
	}
}


