/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.AssociatedPairsEvent;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.Grouping;


/**
 * @author A.G.Docherty
 *
 * Provides resolution by looking for a given number, when it's only possible in two cells of
 * two different rows, and that the four possibilities are on two columns (and form a rectangle), 
 * then that candidate can be excluded for the other cells of the column.
 * The same method applies in the case of two columns, the excluded candidates will then be on the rows.
 * 
 */
public class AssociatedPairsSolver extends AbstractSolver
{
public AssociatedPairsSolver(Puzzle grid)
    {
    super(grid, 100, SolverDifficulty.HARD);
    }

@Override
public int findSolutions() throws IllegalGridException
	{
	found = 0;

	int[][] allFreq = new int[SudokuSolver.GRID_SIZE][SudokuSolver.GRID_SIZE];
	
	// check the rows
	for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
		{
		allFreq[i] = Puzzle.getFrequencyOfPossibilities(hGrid[i]);
		}

	if ( resolveAssocPairs(allFreq, Grouping.ROW) )
		{
		found++;
		}
	else
		{
		// check the columns
		for ( int i = 0; i < SudokuSolver.GRID_SIZE; i++ )
			{
			// create a set of the options column
			Cell[] col = puzzle.getColumn(i);
			allFreq[i] = Puzzle.getFrequencyOfPossibilities(col);
			}
	
		if ( resolveAssocPairs(allFreq, Grouping.COLUMN) )
			found++;
		}
	
	return found;
	}

/**
 * Finds the associated pairs in the rows or columns
 * 
 * @param allFreq - the frequencies of each number in each row or column
 * @param direction - the direction (row or column)
 * @return true if an associated pair was found
 * @throws IllegalGridException
 */
private boolean resolveAssocPairs(int[][] allFreq, Grouping direction) throws IllegalGridException
	{
	// for each group
	for ( int i = 0; i < allFreq.length; i++ )
	    {
	    //for each cell
	    for ( int j = 0; j < allFreq[i].length; j++ )
	        {
	        // if the frequency of this number is 2
	        if ( allFreq[i][j] == 2 )
	        	{
	        	// find a pair of this number
	        	int num = j+1;
	        	int match = getMatchingPairIndex(allFreq, num, i);
	        	while ( match > 0 )
	        		{
	        		// check for a square of numbers
	        		if ( resolveForSquare(direction, i, num, match) )
	        			return true;
	        		
	        		// check for all the numbers in 2 blocks
	        		if ( resolveForBlock(direction, i, num, match) )
	        			return true;
	        		
	        		// get the next pair
	        		match = getMatchingPairIndex(allFreq, num, match);
	        		}
	        	}
	        }
	    }
	
	return false;
	}

/**
 * Resolves a square format of matching numbers 
 * 
 * @param direction the direction (row or column)
 * @param i the index of the first row or column containing the number
 * @param num the matching number
 * @param match the index of the second row or column containing the number
 * @return true if a square of numbers was found and the number was removed from other cells in the puzzle
 * @throws IllegalGridException 
 */
private boolean resolveForSquare(Grouping direction, int i, int num, int match) throws IllegalGridException
	{
	String msg = "Associated Pairs: The number is only possible in two cells of two "+direction.toString().toLowerCase()+"s forming a square.";
	List<Integer> count = new ArrayList<Integer>();
	List<Point> coords = new ArrayList<Point>();
	
	int f = 0;
	
	// find these values in the grid and check they form a square
	for ( int k = 0; k < SudokuSolver.GRID_SIZE; k++ )
	    {
	    if ( direction == Grouping.ROW )
	    	{
			if ( hGrid[i][k].contains(num) )
				{
	    		if ( hGrid[match][k].contains(num) )
	    			{
	    			coords.add(new Point(k, i));
	    			coords.add(new Point(k, match));
	    			count.add(k);
	    			}
				}
	    	}
	    else
	    	{
			if ( hGrid[k][i].contains(num) )
				{
	    		if ( hGrid[k][match].contains(num) )
	    			{
	    			coords.add(new Point(i, k));
	    			coords.add(new Point(match, k));
	    			count.add(k);
	    			}
				}
	    	}
	    }
	
	// if they form a square
	if ( count.size() == 2 )
		{
		for ( Integer lineIndex : count )
	        {
	        // remove this number from the rest of the line at 90 degrees to the found lines
			for ( int k = 0; k < SudokuSolver.GRID_SIZE; k++ )
	            {
	            if ( k == i || k == match )
	            	continue;
	            
	            Cell set;
				if ( direction == Grouping.ROW )
	            	{
	            	set = hGrid[k][lineIndex];
	            	}
	            else
	            	{
	            	set = hGrid[lineIndex][k];
	            	}
            	
        		if ( set.size() > 1 && set.contains(num) )
        			{
        			set.remove(num);
    				firePossibilitiesReduced(hGrid, puzzle.getLocation(set), num, coords, msg, direction);
    				f++;
        			}
	            }
	        }
		}
	
	return (f > 0);
	}

/**
 * Resolves a set of four identical numbers in 2 blocks (2 per block) where the numbers 
 * are in the same two rows or columns but not necessarily both (ie they do not form a square) 
 * 
 * @param direction the direction (row or column)
 * @param i the index of the first row or column containing the number
 * @param num the matching number
 * @param match the index of the second row or column containing the number
 * @return true if four identical numbers in two blocks were found and the number was 
 *              removed from other cells in the puzzle
 * @throws IllegalGridException 
 */
private boolean resolveForBlock(Grouping direction, int i, int num, int match) throws IllegalGridException
	{
	String msg = "Associsted Pairs: The number is only possible in two cells of two "+direction.toString().toLowerCase()+"s in the same blocks.";
	List<Point> coords = new ArrayList<Point>();
	
	// find these values in the grid and check they both in the same block
	for ( int k = 0; k < SudokuSolver.GRID_SIZE; k++ )
	    {
	    if ( direction == Grouping.ROW )
	    	{
			if ( hGrid[i][k].contains(num) )
				{
				int blockIndex = puzzle.getBlockIndex(k, i);
				
				if ( isExistingBlock(coords, blockIndex) )
					continue;
				
				for ( int j = 0; j < SudokuSolver.GRID_SIZE; j++ )
	                {
		    		if ( hGrid[match][j].contains(num) && puzzle.getBlockIndex(j, match) == blockIndex )
		    			{
		    			coords.add(new Point(k, i));
		    			coords.add(new Point(j, match));
		    			}
	                }
				}
	    	}
	    else
	    	{
			if ( hGrid[k][i].contains(num) )
				{
				int blockIndex = puzzle.getBlockIndex(i, k);
				
				if ( isExistingBlock(coords, blockIndex) )
					continue;
				
				for ( int j = 0; j < SudokuSolver.GRID_SIZE; j++ )
	                {
		    		if ( hGrid[j][match].contains(num) && puzzle.getBlockIndex(match, j) == blockIndex )
		    			{
		    			coords.add(new Point(i, k));
		    			coords.add(new Point(match, j));
		    			}
	                }
				}
	    	}
	    }
	
	int f = 0;
	// if they are in the same blocks
	if ( coords.size() == 4 )
		{
		for ( int j = 0; j < coords.size(); j+=2 )
	        {
	        // get the block
	        Point c1 = coords.get(j);
	        Point c2 = coords.get(j+1);
	        Cell[] block = puzzle.getBlock(c1.x, c1.y);

	        Cell s1;
			// get the two sets
	        Cell s2;
           	s1 = hGrid[c1.y][c1.x];
           	s2 = hGrid[c2.y][c2.x];
            
	        // remove this number from the rest of the block
			for ( int k = 0; k < block.length; k++ )
	            {
		        Cell bSet = block[k];
		        
	            if ( bSet == s1 || bSet == s2 )
	            	continue;
	
	            if ( bSet.size() > 1 && bSet.contains(num) )
	            	{
	    			bSet.remove(num);
					firePossibilitiesReduced(hGrid, puzzle.getLocation(bSet), num, coords, msg, direction);
					f ++;
	            	}
	            }
	        }
		}
	
	return f > 0;
	}

/**
 * Finds if any of these coordinates are in this block
 * 
 * @param coords the list of coordinates
 * @param block the block index
 * @return true if any of these coordinates are in this block
 */
private boolean isExistingBlock(List<Point> coords, int block)
	{
	for ( Point p : coords )
	    {
	    if ( puzzle.getBlockIndex(p.x, p.y) == block )
	    	return true;
	    }

	return false;
	}

/**
 * Finds a line which contains this 2 instances of this number
 * 
 * @param allFreq the lines of frequencies 
 * @param num the number to match
 * @param line the index of the line containing the number to match
 * @return the index of another line containing 2 instances of this number
 */
private int getMatchingPairIndex(int[][] allFreq, int num, int line)
	{
	for ( int i = line+1; i < allFreq.length; i++ )
	    {
        // if the cell frequency is 2
        if ( allFreq[i][num-1] == 2 )
        	{
        	return i;
	        }
	    }
	
	return -1;
	}

/**
 * Fires the possibilities reduced event to all listeners
 * @throws IllegalGridException 
 */
private void firePossibilitiesReduced(Cell[][] grid, Point loc, int num, List<Point> from, String msg, Grouping direction) throws IllegalGridException
	{
	if ( !hasSolverListeners() )
		return;
	
	Cell removed = new Cell();
	removed.add(num);
	
	SolutionEvent e = new AssociatedPairsEvent(this, msg, puzzle.clone(), loc, removed, from, direction);
	fireSolutionEvent(e);
	}
}

