/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import static org.junit.Assert.assertTrue;

import java.util.Arrays;

import org.junit.Before;
import org.junit.Test;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.IllegalGridException;

/**
 * @author A.G.Docherty
 * If there is N cells (2,3 or 4) containing N common numbers, then all other possible numbers 
 * for these cells can be excluded.
 */
public class AssociatedPairsSolverTest
{

private Cell[][] grid;
private final String[] args = new String[] { "294135876813746925657298143926813754735964281148572369579621438382459617461387592" };
private SudokuSolver sudoku;
private Puzzle puzzle;

/**
 * @throws java.lang.Exception
 */
@Before
public void setUp() throws Exception
	{
	sudoku = new SudokuSolver(args);
	puzzle = sudoku.getPuzzle();
	grid = puzzle.getGrid();
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.ZeroChoiceSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsRowsSq() throws IllegalGridException
	{
	// Assoc Pair Number = 8
	puzzle.set(0, 1, new Cell(Arrays.asList(2, 4, 3)));
	puzzle.set(2, 1, new Cell(Arrays.asList(8, 4, 5)));
	puzzle.set(4, 1, new Cell(Arrays.asList(2, 3)));
	puzzle.set(6, 1, new Cell(Arrays.asList(4, 6)));
	puzzle.set(7, 1, new Cell(Arrays.asList(4, 7)));
	puzzle.set(8, 1, new Cell(Arrays.asList(4, 8, 6, 5)));

	puzzle.set(2, 5, new Cell(Arrays.asList(3, 8, 7)));
	puzzle.set(6, 5, new Cell(Arrays.asList(3, 8, 6)));
	puzzle.set(8, 5, new Cell(Arrays.asList(3, 8, 5)));
	puzzle.set(2, 6, new Cell(Arrays.asList(8, 9)));
	puzzle.set(8, 6, new Cell(Arrays.asList(8, 9)));
	
//	sudoku.printDetail();
	
	assertTrue(grid[1][0].size() > 1);
	assertTrue(grid[1][2].size() > 1);
	assertTrue(grid[1][4].size() > 1);
	assertTrue(grid[1][6].size() > 1);
	assertTrue(grid[1][7].size() > 1);
	assertTrue(grid[1][8].size() > 1);

	assertTrue(grid[5][2].size() > 1);
	assertTrue(grid[5][6].size() > 1);
	assertTrue(grid[5][8].size() > 1);
	assertTrue(grid[6][2].size() > 1);
	assertTrue(grid[6][8].size() > 1);
	
	AssociatedPairsSolver solver = new AssociatedPairsSolver(puzzle);
	solver.findSolutions();

//	sudoku.printDetail();

	assertTrue(grid[1][0].size() == 3);
	assertTrue(grid[1][2].size() == 3);
	assertTrue(grid[1][4].size() == 2);
	assertTrue(grid[1][6].size() == 2);
	assertTrue(grid[1][7].size() == 2);
	assertTrue(grid[1][8].size() == 4);
	
	assertTrue(grid[5][2].size() == 2);
	assertTrue(grid[5][6].size() == 3);
	assertTrue(grid[5][8].size() == 2);
	assertTrue(grid[6][2].size() == 2);
	assertTrue(grid[6][8].size() == 2);
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.ZeroChoiceSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsColumnsSq() throws IllegalGridException
	{
	// Assoc Pair Number = 8
	puzzle.set(1, 0, new Cell(Arrays.asList(2, 4, 3)));
	puzzle.set(1, 2, new Cell(Arrays.asList(8, 4, 5)));
	puzzle.set(1, 4, new Cell(Arrays.asList(2, 3)));
	puzzle.set(1, 6, new Cell(Arrays.asList(4, 6)));
	puzzle.set(1, 7, new Cell(Arrays.asList(4, 7)));
	puzzle.set(1, 8, new Cell(Arrays.asList(4, 8, 6, 5)));
                  
	puzzle.set(6, 2, new Cell(Arrays.asList(3, 8, 7)));
	puzzle.set(6, 6, new Cell(Arrays.asList(3, 8, 6)));
	puzzle.set(6, 8, new Cell(Arrays.asList(3, 8, 5)));
	puzzle.set(5, 2, new Cell(Arrays.asList(8, 9)));
	puzzle.set(5, 8, new Cell(Arrays.asList(8, 9)));

//	sudoku.printDetail();
	
	assertTrue(grid[0][1].size() > 1);
	assertTrue(grid[2][1].size() > 1);
	assertTrue(grid[4][1].size() > 1);
	assertTrue(grid[6][1].size() > 1);
	assertTrue(grid[7][1].size() > 1);
	assertTrue(grid[8][1].size() > 1);
                      
	assertTrue(grid[2][6].size() > 1);
	assertTrue(grid[6][6].size() > 1);
	assertTrue(grid[8][6].size() > 1);
	assertTrue(grid[2][5].size() > 1);
	assertTrue(grid[8][5].size() > 1);
	
	AssociatedPairsSolver solver = new AssociatedPairsSolver(puzzle);
	solver.findSolutions();

//	sudoku.printDetail();

	assertTrue(grid[0][1].size() == 3);
	assertTrue(grid[2][1].size() == 3);
	assertTrue(grid[4][1].size() == 2);
	assertTrue(grid[6][1].size() == 2);
	assertTrue(grid[7][1].size() == 2);
	assertTrue(grid[8][1].size() == 4);
	                  
	assertTrue(grid[2][6].size() == 2);
	assertTrue(grid[6][6].size() == 3);
	assertTrue(grid[8][6].size() == 2);
	assertTrue(grid[2][5].size() == 2);
	assertTrue(grid[8][5].size() == 2);
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.ZeroChoiceSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsRowsBlock() throws IllegalGridException
	{
	// Assoc Pair Number = 8
	puzzle.set(6, 0, new Cell(Arrays.asList(4)));
	puzzle.set(0, 1, new Cell(Arrays.asList(2, 4, 3)));
	puzzle.set(1, 1, new Cell(Arrays.asList(8, 4, 5)));
	puzzle.set(4, 1, new Cell(Arrays.asList(2, 3)));
	puzzle.set(6, 1, new Cell(Arrays.asList(4, 6)));
	puzzle.set(7, 1, new Cell(Arrays.asList(4, 7)));
	puzzle.set(8, 1, new Cell(Arrays.asList(4, 8, 6, 5)));

	puzzle.set(2, 2, new Cell(Arrays.asList(3, 7, 8)));
	puzzle.set(6, 2, new Cell(Arrays.asList(3, 8, 6)));
	puzzle.set(8, 2, new Cell(Arrays.asList(3, 8, 5)));
	puzzle.set(2, 0, new Cell(Arrays.asList(8, 9)));
	puzzle.set(8, 0, new Cell(Arrays.asList(8, 9)));

//	sudoku.printDetail();

	assertTrue(grid[1][0].size() > 1);
	assertTrue(grid[1][1].size() > 1);
	assertTrue(grid[1][4].size() > 1);
	assertTrue(grid[1][6].size() > 1);
	assertTrue(grid[1][7].size() > 1);
	assertTrue(grid[1][8].size() > 1);

	assertTrue(grid[2][2].size() > 1);
	assertTrue(grid[2][6].size() > 1);
	assertTrue(grid[2][8].size() > 1);
	assertTrue(grid[0][2].size() > 1);
	assertTrue(grid[0][8].size() > 1);
	
	AssociatedPairsSolver solver = new AssociatedPairsSolver(puzzle);
	solver.findSolutions();

//	sudoku.printDetail();

	assertTrue(grid[1][0].size() == 3);
	assertTrue(grid[1][1].size() == 3);
	assertTrue(grid[1][4].size() == 2);
	assertTrue(grid[1][6].size() == 2);
	assertTrue(grid[1][7].size() == 2);
	assertTrue(grid[1][8].size() == 4);
	
	assertTrue(grid[2][2].size() == 2);
	assertTrue(grid[2][6].size() == 2);
	assertTrue(grid[2][8].size() == 2);
	assertTrue(grid[0][2].size() == 2);
	assertTrue(grid[0][8].size() == 2);
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.ZeroChoiceSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsColumnsBlock() throws IllegalGridException
	{
	// Assoc Pair Number = 8
	puzzle.set(0, 1, new Cell(Arrays.asList(4)));
	puzzle.set(1, 0, new Cell(Arrays.asList(2, 4, 3)));
	puzzle.set(1, 1, new Cell(Arrays.asList(8, 4, 5)));
	puzzle.set(1, 4, new Cell(Arrays.asList(2, 3)));
	puzzle.set(1, 6, new Cell(Arrays.asList(4, 6)));
	puzzle.set(1, 7, new Cell(Arrays.asList(4, 7)));
	puzzle.set(1, 8, new Cell(Arrays.asList(4, 8, 6, 5)));
                  
	puzzle.set(2, 2, new Cell(Arrays.asList(3, 7, 8)));
	puzzle.set(2, 6, new Cell(Arrays.asList(3, 8, 6)));
	puzzle.set(2, 8, new Cell(Arrays.asList(3, 8, 5)));
	puzzle.set(0, 2, new Cell(Arrays.asList(8, 9)));
	puzzle.set(0, 8, new Cell(Arrays.asList(8, 9)));

//	sudoku.printDetail();

	assertTrue(grid[0][1].size() > 1);
	assertTrue(grid[1][1].size() > 1);
	assertTrue(grid[4][1].size() > 1);
	assertTrue(grid[6][1].size() > 1);
	assertTrue(grid[7][1].size() > 1);
	assertTrue(grid[8][1].size() > 1);
                      
	assertTrue(grid[2][2].size() > 1);
	assertTrue(grid[6][2].size() > 1);
	assertTrue(grid[8][2].size() > 1);
	assertTrue(grid[2][0].size() > 1);
	assertTrue(grid[8][0].size() > 1);
	
	AssociatedPairsSolver solver = new AssociatedPairsSolver(puzzle);
	solver.findSolutions();

//	sudoku.printDetail();

	assertTrue(grid[0][1].size() == 3);
	assertTrue(grid[1][1].size() == 3);
	assertTrue(grid[4][1].size() == 2);
	assertTrue(grid[6][1].size() == 2);
	assertTrue(grid[7][1].size() == 2);
	assertTrue(grid[8][1].size() == 4);
	                  
	assertTrue(grid[2][2].size() == 2);
	assertTrue(grid[6][2].size() == 2);
	assertTrue(grid[8][2].size() == 2);
	assertTrue(grid[2][0].size() == 2);
	assertTrue(grid[8][0].size() == 2);

	}

}
