/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.FixedLineEvent;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.Grouping;

/**
 * @author A.G.Docherty
 * 
 * Provides resolution by looking for numbers in a block that have
 * to be on particular row or column and therefore can't be in that row
 * (or column) in another block.
 * 
 */
public class FixedLineSolver extends AbstractSolver
{
public FixedLineSolver(Puzzle grid)
    {
    super(grid, 4, SolverDifficulty.BEGINNER);
    }

@Override
public int findSolutions() throws IllegalGridException
	{
	found = 0;

	for ( int i = 0; i < SudokuSolver.SUB_GRID_SIZE && found == 0; i++ )
		{
		int j = i * SudokuSolver.SUB_GRID_SIZE;
		List<Cell[]> blocks = Arrays.asList(puzzle.getBlock(j),
				puzzle.getBlock(j + 1), puzzle.getBlock(j + 2));

		for ( int k = 0; k < SudokuSolver.SUB_GRID_SIZE && found == 0; k++ )
			{
			found += resolveHLines(k, blocks);
			}
		}

	for ( int i = 0; i < SudokuSolver.SUB_GRID_SIZE && found == 0; i++ )
		{
		List<Cell[]> blocks = Arrays.asList(puzzle.getBlock(i),
				puzzle.getBlock(i + SudokuSolver.SUB_GRID_SIZE), puzzle.getBlock(i
		                + (SudokuSolver.SUB_GRID_SIZE * 2)));

		for ( int k = 0; k < SudokuSolver.SUB_GRID_SIZE && found == 0; k++ )
			{
			found += resolveVLines(k, blocks);
			}
		}

	return found;
	}

private int resolveHLines(int row, List<Cell[]> blocks) throws IllegalGridException
	{
	String msg = "Fixed Row: The value has to be in the referenced row of this block";
	int f = 0;
	Cell total = new Cell();

	for ( int i = 0; i < blocks.size(); i++ )
		{
		total.clear();

		// get all the number in this row for this block
		for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
			{
			Cell set = blocks.get(i)[(row * SudokuSolver.SUB_GRID_SIZE) + j];

			total.addAll(set);
			}

		// find which numbers are only in this row in this block
		// by removing numbers that maybe in a different row in this block
		for ( int r = 0; r < SudokuSolver.SUB_GRID_SIZE; r++ )
			{
			if ( r == row )
				continue;

			for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
				{
				Cell set = blocks.get(i)[(r * SudokuSolver.SUB_GRID_SIZE) + j];
				total.removeAll(set);
				}
			}

		// if there's any numbers left remove them from the rest of this line
		for ( int b = 0; b < blocks.size(); b++ )
			{
			if ( b == i )
				continue;

			for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
				{
				Cell set = blocks.get(b)[(row * SudokuSolver.SUB_GRID_SIZE)+ j];

				if ( set.size() == 1 )
					continue;
					
				Cell removed = removeAllFromSet(set, total);

				if ( removed.size() > 0 )
					{
					// find which sets these remaining numbers are in
					List<Point> coords = new ArrayList<Point>();
					for ( int r = 0; r < SudokuSolver.SUB_GRID_SIZE; r++ )
						{
						for ( int k = 0; k < SudokuSolver.SUB_GRID_SIZE; k++ )
							{
							Cell s = blocks.get(i)[r + (k * SudokuSolver.SUB_GRID_SIZE)];
							
							// if this set contains any of the numbers in total then keep a record of these coordinates
							if ( hasMatchingNumber(removed, s) )
								coords.add(puzzle.getLocation(s));
							}
						}
					
					f += removed.size();
					firePossibilitiesReduced(hGrid, puzzle.getLocation(set), removed, coords, msg, Grouping.ROW);
					}
				}
			}
		}

	return f;
	}

private int resolveVLines(int col, List<Cell[]> blocks) throws IllegalGridException
	{
	String msg = "Fixed Column: The value has to be in the referenced column of this block";
	int f = 0;
	Cell total = new Cell();

	for ( int i = 0; i < blocks.size(); i++ )
		{
		total.clear();

		// get all the number in this blocks column
		for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
			{
			Cell set = blocks.get(i)[col + (j * SudokuSolver.SUB_GRID_SIZE)];

			total.addAll(set);
			}

		// find which numbers are only in this row in this block
		for ( int c = 0; c < SudokuSolver.SUB_GRID_SIZE; c++ )
			{
			if ( c == col )
				continue;

			for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
				{
				Cell set = blocks.get(i)[c + (j * SudokuSolver.SUB_GRID_SIZE)];
				total.removeAll(set);
				}
			}

		// if there's any numbers left remove them from the rest of this line
		for ( int b = 0; b < blocks.size(); b++ )
			{
			if ( b == i )
				continue;

			for ( int j = 0; j < SudokuSolver.SUB_GRID_SIZE; j++ )
				{
				Cell set = blocks.get(b)[col + (j * SudokuSolver.SUB_GRID_SIZE)];
				
				if ( set.size() == 1 )
					continue;
					
				Cell removed = removeAllFromSet(set, total);

				if ( removed.size() > 0 )
					{
					// find which sets these remaining numbers are in
					List<Point> coords = new ArrayList<Point>();
					for ( int c = 0; c < SudokuSolver.SUB_GRID_SIZE; c++ )
						{
						for ( int k = 0; k < SudokuSolver.SUB_GRID_SIZE; k++ )
							{
							Cell s = blocks.get(i)[(c * SudokuSolver.SUB_GRID_SIZE) + k];
							
							// if this set contains any of the numbers in total then keep a record of these coordinates
							if ( hasMatchingNumber(removed, s) )
								coords.add(puzzle.getLocation(s));
							}
						}

					f += removed.size();
					firePossibilitiesReduced(hGrid, puzzle.getLocation(set), removed, coords, msg, Grouping.COLUMN);
					}
				}
			}
		}

	return f;
	}

/**
 * Returns true if at least one number is in both sets
 * 
 * @param total
 * @param set
 * @return
 */
private boolean hasMatchingNumber(Cell total, Cell set)
	{
	for ( Integer i : total )
	    {
	    if ( set.contains(i) )
	    	return true;
	    }

	return false;
	}


/**
 * fires the possibilities reduced event to all listeners
 * @throws IllegalGridException 
 */
private void firePossibilitiesReduced(Cell[][] grid, Point loc, Cell removed, List<Point> from, String msg, Grouping grouping) throws IllegalGridException
	{
	if ( !hasSolverListeners() )
		return;
	
	SolutionEvent e = new FixedLineEvent(this, msg, puzzle.clone(), loc, removed, from, grouping);
	fireSolutionEvent(e);
	}

}
