/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.HiddenGroupsEvent;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.Grouping;

/**
 * @author A.G.Docherty
 *
 * Provides resolution by looking for numbers which only occur n times and whose occurrences are in
 * the same n cells. All other numbers in those cells can then be removed. 
 *  
 */
public class HiddenGroupsSolver extends AbstractSolver
{
private final String[] type = new String[] {"Singles", "Pairs", "Triplets", "Quadruplets"};

public HiddenGroupsSolver(Puzzle grid)
    {
    super(grid, 30, SolverDifficulty.MEDIUM);
    }

@Override
public int findSolutions() throws IllegalGridException
	{
	found = 0;

	for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
		{
		found += resolveHiddenGroups(hGrid[i], Grouping.ROW);
		}

	for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
		{
		// create a set of the options column
		Cell[] col = puzzle.getColumn(i);

		found += resolveHiddenGroups(col, Grouping.COLUMN);
		}

	for ( int i = 0; i < SudokuSolver.GRID_SIZE && found == 0; i++ )
		{
		// create a set of the options block
		Cell[] col = puzzle.getBlock(i);

		found += resolveHiddenGroups(col, Grouping.BLOCK);
		}

	return found;
	}

private int resolveHiddenGroups(Cell[] sets, Grouping direction) throws IllegalGridException
	{
	int f = 0;
	int[] occur = Puzzle.getFrequencyOfPossibilities(sets);

	List<Integer> matches = new ArrayList<Integer>();
	
	for ( int n = 2; n < 5 && f == 0; n++ )
	    {
		matches.clear();
		List<Cell> locs = new ArrayList<Cell>();
		
		// find all numbers that only occurs n times
		for ( int i = 0; i < occur.length; i++ )
			{
			if ( occur[i] == n )
				{
				// find all locations this number is in
				locs.add(getAllLocations(sets, i+1));
				}
			}
		
		// find matching sets
		do 	{
			Cell match = getMatchingSets(locs, n);
			
			if ( match == null )
				break;
			
			// remove any extra value from the sets
			f += removeExtraValues(sets, match, occur, direction);
			}
		while ( true);
		}
	
	return f;
	}
	

/**
 * Removes any values that are not common to all match indices
 * 
 * @param sets
 * @param match
 * @throws IllegalGridException 
 */
private int removeExtraValues(Cell[] line, Cell match, int[] freq, Grouping direction) throws IllegalGridException
	{
	String msg = "Hidden Group: Eliminate extra possible numbers from these embedded "+type[match.size()-1];
	
    Cell[] sets = new Cell[match.size()];
	List<Point> coords = new ArrayList<Point>();
	int j = 0;
	
	for ( Integer i : match )
	    {
	    sets[j++] = line[i];
		coords.add(puzzle.getLocation(line[i]));
	    }
	
	int[] occurs = Puzzle.getFrequencyOfPossibilities(sets);
	
	// get the numbers that occur n times and have an overall frequency of n
	Cell common = new Cell();
	for ( int i = 0; i < occurs.length; i++ )
	    {
	    if ( occurs[i] == match.size() && freq[i] == match.size() )
	    	common.add(i+1);
	    }
	
	int f = 0;
	
	// retain only the common values.
	for ( Integer i : match )
	    {
	    Cell removed = retainAllInSet(line[i], common);
	    
	    if ( removed.size()> 0 )
	    	{
	    	f += removed.size();
			firePossibilitiesReduced(hGrid, common, puzzle.getLocation(line[i]), removed, coords, msg, direction);
	    	}
	    }
	
	return f;
	}

/**
 * Gets matching sets that occur n times and remove them from the list
 * 
 * @param locs
 * @param n
 * @return
 */
private Cell getMatchingSets(List<Cell> locs, int n)
	{
	while ( locs.size() >= n )
	    {
	    int match = 1;
	    
	    for ( int j = 1; j < locs.size(); j++ )
	        {
	        if ( locs.get(j).equals(locs.get(0)) )
	        	{
	        	match++;
	        	}
	        }
	    
	    if ( match == n )
	    	{
	    	Cell set = locs.get(0);
	    	
	    	while ( locs.contains(set ))
	    		locs.remove(set);
	    	
	    	return set;
	    	}
	    
    	locs.remove(0);
	    }
	
	return null;
	}

/**
 * gets the zero based index of the sets that contain this number
 * 
 * @param sets
 * @param num
 * @return
 */
private Cell getAllLocations(Cell[] sets, int num)
	{
	Cell in = new Cell();
	
	for ( int i = 0; i < sets.length; i++ )
	    {
	    if ( sets[i].contains(num) )
	    	in.add(i);
	    }
	
	return in;
	}


/**
 * fires the possibilities reduced event to all listeners
 * @throws IllegalGridException 
 */
private void firePossibilitiesReduced(Cell[][] grid, Cell ref, Point loc, Cell removed, List<Point> from, String msg, Grouping direction) throws IllegalGridException
	{
	if ( !hasSolverListeners() )
		return;
	
	SolutionEvent e = new HiddenGroupsEvent(this, msg, puzzle.clone(), ref, loc, removed, from, direction);
	fireSolutionEvent(e);
	}


}
