/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import java.awt.Point;
import java.util.Arrays;
import java.util.List;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.IllegalGridException;
import uk.co.keang.sudoku.event.KnownValueEvent;
import uk.co.keang.sudoku.event.SolutionEvent;
import uk.co.keang.sudoku.event.SolutionEvent.Grouping;


/**
 * Provides resolution by simply looking along row, column and block and for each cell
 * removing all known values in the row, column or block.
 * 
 * Because of the basic nature of this solver scans the whole grid before returning   
 *
 * @author A.G.Docherty
 *
 */
public class KnownValueSolver extends AbstractSolver
{
public KnownValueSolver(Puzzle grid)
    {
    super(grid, 2, SolverDifficulty.BEGINNER);
    }

@Override
public int findSolutions() throws IllegalGridException
	{
	found = 0;

	for ( int y = 0; y < SudokuSolver.GRID_SIZE; y++ )
		{
		for ( int x = 0; x < SudokuSolver.GRID_SIZE; x++ )
			{
			if ( hGrid[y][x].size() > 1 )
				{
				// remove known values from all possible values for this cell
				if ( !removeValsInRow(x, y) )
					{
					if ( !removeValsInColumn(x, y) )
						{
						removeValsInBlock(x, y);
						}
					}
				}
			}
		}

	return found;
	}

/**
 * Remove known numbers in this block from the possible numbers of this cell 
 * 
 * @param x the x coordinate (0 based) of the cell to resolve 
 * @param y the y coordinate (0 based) of the cell to resolve
 * @return true if the cell has been resolved
 * @throws IllegalGridException 
 */
private boolean removeValsInBlock(int x, int y) throws IllegalGridException
	{
	// remove values already in block
	Cell[] b = puzzle.getBlock(x, y);
	String msg = "Known Values: Eliminate known values.";
	return removeValsInLine(x, y, b, msg, Grouping.BLOCK);
	}

/**
 * Remove known numbers in this column from the possible numbers of this cell 
 * 
 * @param x the x coordinate (0 based) of the cell to resolve 
 * @param y the y coordinate (0 based) of the cell to resolve
 * @return true if the cell has been resolved
 * @throws IllegalGridException 
 */
private boolean removeValsInColumn(int x, int y) throws IllegalGridException
	{
	// remove values already in col
	Cell[] col = puzzle.getColumn(x);
	String msg = "Known Values: Eliminate known values.";
	return removeValsInLine(x, y, col, msg, Grouping.COLUMN);
	}

/**
 * Remove known numbers in this row from the possible numbers of this cell 
 * 
 * @param x the x coordinate (0 based) of the cell to resolve 
 * @param y the y coordinate (0 based) of the cell to resolve
 * @return true if the cell has been resolved
 * @throws IllegalGridException 
 */
private boolean removeValsInRow(int x, int y) throws IllegalGridException
	{
	String msg = "Known Values: Eliminate known values.";
	return removeValsInLine(x, y, hGrid[y], msg, Grouping.ROW);
	}

/**
 * Removes the values that match in this row, column or block
 * 
 * @param x the x coordinate (0 based) of the cell to resolve 
 * @param y the y coordinate (0 based) of the cell to resolve
 * @param line the grouping to remove values from
 * @param msg the msg to use to describe this solution
 * @param direction the grouping
 * @return true if the cell has been resolved
 * @throws IllegalGridException 
 */
private boolean removeValsInLine(int x, int y, Cell[] line, String msg, Grouping direction) throws IllegalGridException
	{
	Cell opt = hGrid[y][x];

	// remove values already in row
	for ( Cell sq : line )
		{
		if ( sq.size() == 1 )
			{
			Cell removed = removeAllFromSet(opt, sq);

			if ( removed.size() > 0 )
				{
				firePossibilitiesReduced(hGrid, puzzle.getLocation(opt), removed, Arrays.asList(puzzle.getLocation(sq)), msg, direction);
				}
			
			if ( opt.size() == 1 )
				{
				// use this value
				found++;
				return true;
				}
			}
		}

	return false;
	}

/**
 * Fires the possibilities reduced event to all listeners
 * 
 * @throws IllegalGridException 
 */
private void firePossibilitiesReduced(Cell[][] grid, Point loc, Cell removed, List<Point> from, String msg, Grouping direction) throws IllegalGridException
	{
	if ( !hasSolverListeners() )
		return;
	
	SolutionEvent e = new KnownValueEvent(this, msg, puzzle.clone(), loc, removed, from, direction);
	fireSolutionEvent(e);
	}
}

