/**
 *  Copyright (c) 2010 A.G.Docherty. All Rights Reserved.
 */

package uk.co.keang.sudoku.solver;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.junit.Before;
import org.junit.Test;

import uk.co.keang.sudoku.Puzzle;
import uk.co.keang.sudoku.Puzzle.Cell;
import uk.co.keang.sudoku.SudokuSolver;
import uk.co.keang.sudoku.event.IllegalGridException;

/**
 * @author A.G.Docherty
 *
 */
public class KnownValueSolverTest
{
private Cell[][] grid;
private final String[] args = new String[] { "294135876813746925657298143926813754735964281148572369579621438382459617461387592" };
private SudokuSolver sudoku;
private Puzzle puzzle;

/**
 * @throws java.lang.Exception
 */
@Before
public void setUp() throws Exception
	{
	sudoku = new SudokuSolver(args);
	puzzle = sudoku.getPuzzle();
	grid = puzzle.getGrid();
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.FixedLineSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsRows() throws IllegalGridException
	{
	for ( int i = 0; i < grid.length-3; i++ )
	    {
	    puzzle.set(i, i, -1);
	    }
	
	for ( int i = grid.length-3; i < grid.length; i++ )
	    {
	    for ( int j = 0; j < grid[i].length; j++ )
	        {
		    puzzle.set(j, i, -1);
	        }
	    }

//	sudoku.printDetail();
	
	// test changes have been done
	for ( int i = 0; i < grid.length-3; i++ )
	    {
	    assertTrue(grid[i][i].size() > 1);
	    }
	for ( int i = grid.length-3; i < grid.length; i++ )
	    {
	    for ( int j = 0; j < grid[i].length; j++ )
	        {
		    assertTrue(grid[i][j].size() > 1);
	        }
	    }
	
	KnownValueSolver solver = new KnownValueSolver(puzzle);
	solver.findSolutions();
	
//	sudoku.printDetail();
	
	// test cells have been resolved
	for ( int i = 0; i < grid.length-3; i++ )
	    {
	    assertTrue(grid[i][i].size() == 1);
	    }
	for ( int i = grid.length-3; i < grid.length; i++ )
	    {
	    for ( int j = 0; j < grid[i].length; j++ )
	        {
		    assertTrue(grid[i][j].size() == 3);
	        }
	    }
	
	assertFalse(puzzle.isComplete());
	assertTrue(puzzle.isLegal());
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.FixedLineSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsColumns() throws IllegalGridException
	{
	for ( int i = 0; i < grid.length-3; i++ )
	    {
	    puzzle.set(i, i, -1);
	    }
	
	for ( int i = grid.length-3; i < grid.length; i++ )
	    {
	    for ( int j = 0; j < grid[i].length; j++ )
	        {
		    puzzle.set(i, j, -1);
	        }
	    }

//	sudoku.printDetail();
	
	// test changes have been done
	for ( int i = 0; i < grid.length-3; i++ )
	    {
	    assertTrue(grid[i][i].size() > 1);
	    }
	for ( int i = grid.length-3; i < grid.length; i++ )
	    {
	    for ( int j = 0; j < grid[i].length; j++ )
	        {
		    assertTrue(grid[j][i].size() > 1);
	        }
	    }
	
	KnownValueSolver solver = new KnownValueSolver(puzzle);
	solver.findSolutions();
	
//	sudoku.printDetail();
	
	// test cells have been resolved
	for ( int i = 0; i < grid.length-3; i++ )
	    {
	    assertTrue(grid[i][i].size() == 1);
	    }
	for ( int i = grid.length-3; i < grid.length; i++ )
	    {
	    for ( int j = 0; j < grid[i].length; j++ )
	        {
		    assertTrue(grid[j][i].size() == 3);
	        }
	    }
	
	assertFalse(puzzle.isComplete());
	assertTrue(puzzle.isLegal());
	}

/**
 * Test method for {@link uk.co.keang.sudoku.solver.FixedLineSolver#findSolutions()}.
 * @throws IllegalGridException 
 */
@Test
public void testFindSolutionsBlock() throws IllegalGridException
	{
    puzzle.set(1, 1, -1);
	
	for ( int i = grid.length-3; i < grid.length; i++ )
	    {
	    for ( int j = 0; j < grid[i].length; j++ )
	        {
		    puzzle.set(i, j, -1);
		    puzzle.set(j, i, -1);
	        }
	    }

//	sudoku.printDetail();
	
	// test changes have been done
    assertTrue(grid[1][1].size() > 1);
	for ( int i = grid.length-3; i < grid.length; i++ )
	    {
	    for ( int j = 0; j < grid[i].length; j++ )
	        {
		    assertTrue(grid[j][i].size() > 1);
		    assertTrue(grid[i][j].size() > 1);
	        }
	    }
	
	KnownValueSolver solver = new KnownValueSolver(puzzle);
	solver.findSolutions();
	
//	sudoku.printDetail();
	
	// test cells have been resolved
    assertTrue(grid[1][1].size() == 1);
	for ( int i = grid.length-3; i < grid.length; i++ )
	    {
	    for ( int j = 0; j < grid[i].length-1; j++ )
	        {
	        if ( i > (SudokuSolver.GRID_SIZE-SudokuSolver.SUB_GRID_SIZE-1) && j > (SudokuSolver.GRID_SIZE-SudokuSolver.SUB_GRID_SIZE-1))
			    assertTrue(grid[j][i].size() == 9);
	        else
	        	{
			    assertTrue(grid[j][i].size() == 3);
			    assertTrue(grid[i][j].size() == 3);
	        	}
	        }
	    }
	
	assertFalse(puzzle.isComplete());
	assertTrue(puzzle.isLegal());
	}

}
